# -*- coding: utf-8 -*-
"""
Created on Fri Mar  22 2024

@author: Riccardo
"""

import numpy as np
import casadi as cas
import rockit
import invariants_py.dynamics_invariants as dynamics
from invariants_py.forward_kinematics import forward_kinematics
from invariants_py.ocp_helper import check_solver, tril_vec, tril_vec_no_diag, diffR, diag

class OCP_gen_pose_jointlim:

    def __init__(self, path_to_urdf, window_len = 100, bool_unsigned_invariants = False, w_pos = 1, w_rot = 1, max_iters = 300, fatrop_solver = False, nb_joints = 6, root = 'base_link', tip = 'tool0'):
        fatrop_solver = check_solver(fatrop_solver)       
        #%% Create decision variables and parameters for the optimization problem
        
        ocp = rockit.Ocp(T=1.0)
        
        # Define system states X (unknown object pose + moving frame pose at every time step) 
        p_obj = ocp.state(3) # object position
        R_obj_x = ocp.state(3,1) # object orientation
        R_obj_y = ocp.state(3,1) # object orientation
        R_obj_z = ocp.state(3,1) # object orientation
        R_obj = cas.horzcat(R_obj_x,R_obj_y,R_obj_z)
        R_t_x = ocp.state(3,1) # translational Frenet-Serret frame
        R_t_y = ocp.state(3,1) # translational Frenet-Serret frame
        R_t_z = ocp.state(3,1) # translational Frenet-Serret frame
        R_t = cas.horzcat(R_t_x,R_t_y,R_t_z)
        R_r_x = ocp.state(3,1) # rotational Frenet-Serret frame
        R_r_y = ocp.state(3,1) # rotational Frenet-Serret frame
        R_r_z = ocp.state(3,1) # rotational Frenet-Serret frame
        R_r = cas.horzcat(R_r_x,R_r_y,R_r_z)
        q = ocp.state(nb_joints)

        # Define system controls (invariants at every time step)
        U = ocp.control(6)
        qdot = ocp.control(nb_joints)

        # Define system parameters P (known values in optimization that need to be set right before solving)
        h = ocp.parameter(1)
        
        # Boundary values
        R_t_start = ocp.parameter(3,3)
        R_t_end = ocp.parameter(3,3)
        R_r_start = ocp.parameter(3,3)
        R_r_end = ocp.parameter(3,3)
        p_obj_start = ocp.parameter(3)
        p_obj_end = ocp.parameter(3)
        R_obj_start = ocp.parameter(3,3)
        R_obj_end = ocp.parameter(3,3)
        q_lim = ocp.parameter(nb_joints)
        
        U_demo = ocp.parameter(6,grid='control',include_last=True) # model invariants
        
        w_invars = ocp.parameter(6,grid='control',include_last=True) # weights for invariants
                
        #%% Specifying the constraints
        
        # Constrain rotation matrices to be orthogonal (only needed for one timestep, property is propagated by integrator)
        ocp.subject_to(ocp.at_t0(tril_vec(R_t.T @ R_t - np.eye(3))==0.))
        ocp.subject_to(ocp.at_t0(tril_vec(R_r.T @ R_r - np.eye(3))==0.))
        ocp.subject_to(ocp.at_t0(tril_vec(R_obj.T @ R_obj - np.eye(3))==0.))
        
        # Boundary constraints
        ocp.subject_to(ocp.at_t0(tril_vec_no_diag(R_t.T @ R_t_start - np.eye(3))) == 0.)
        ocp.subject_to(ocp.at_t0(tril_vec_no_diag(R_r.T @ R_r_start - np.eye(3)) == 0.))
        ocp.subject_to(ocp.at_tf(tril_vec_no_diag(R_t.T @ R_t_end - np.eye(3)) == 0.))
        ocp.subject_to(ocp.at_tf(tril_vec_no_diag(R_r.T @ R_r_end - np.eye(3)) == 0.))
        ocp.subject_to(ocp.at_t0(p_obj == p_obj_start))
        ocp.subject_to(ocp.at_tf(p_obj == p_obj_end))
        ocp.subject_to(ocp.at_t0(tril_vec_no_diag(R_obj.T @ R_obj_start - np.eye(3)) == 0.))
        ocp.subject_to(ocp.at_tf(tril_vec_no_diag(R_obj.T @ R_obj_end - np.eye(3))==0.))
        for i in range(nb_joints):
            # ocp.subject_to(-q_lim[i] <= (q[i] <= q_lim[i])) # This constraint definition does not work with fatrop, yet
            ocp.subject_to(-q_lim[i] - q[i] <= 0 )
            ocp.subject_to(q[i] - q_lim[i] <= 0)

        # Dynamic constraints
        (R_t_plus1, p_obj_plus1) = dynamics.vector_invariants_position(R_t, p_obj, U[3:], h)
        (R_r_plus1, R_obj_plus1) = dynamics.dyn_vector_invariants_rotation(R_r, R_obj, U[:3], h)
        # Integrate current state to obtain next state (next rotation and position)
        ocp.set_next(p_obj,p_obj_plus1)
        ocp.set_next(R_obj_x,R_obj_plus1[:,0])
        ocp.set_next(R_obj_y,R_obj_plus1[:,1])
        ocp.set_next(R_obj_z,R_obj_plus1[:,2])
        ocp.set_next(R_t_x,R_t_plus1[:,0])
        ocp.set_next(R_t_y,R_t_plus1[:,1])
        ocp.set_next(R_t_z,R_t_plus1[:,2])
        ocp.set_next(R_r_x,R_r_plus1[:,0])
        ocp.set_next(R_r_y,R_r_plus1[:,1])
        ocp.set_next(R_r_z,R_r_plus1[:,2])
        ocp.set_next(q,qdot)

        # Forward kinematics
        p_obj_fwkin, R_obj_fwkin = forward_kinematics(q,path_to_urdf,root,tip)
            
        # Lower bounds on controls
        # if bool_unsigned_invariants:
        #     ocp.subject_to(U[0,:]>=0) # lower bounds on control
        #     ocp.subject_to(U[1,:]>=0) # lower bounds on control
            
        #%% Specifying the objective
        # Fitting constraint to remain close to measurements
        objective_inv = ocp.sum(1/window_len*cas.dot(w_invars*(U - U_demo),w_invars*(U - U_demo)),include_last=True)

        # Objective for joint limits
        e_pos = cas.dot(p_obj_fwkin - p_obj,p_obj_fwkin - p_obj)
        e_rot = cas.dot(R_obj.T @ R_obj_fwkin - np.eye(3),R_obj.T @ R_obj_fwkin - np.eye(3))
        objective_jointlim = ocp.sum(e_pos + e_rot + 0.001*cas.dot(qdot,qdot),include_last = True)

        objective = ocp.sum(objective_inv + objective_jointlim, include_last = True)

        #%% Define solver and save variables
        ocp.add_objective(objective)
        if fatrop_solver:
            ocp.method(rockit.external_method('fatrop' , N=window_len-1))
            # ocp._method.set_name("generation_pose")
            # TEMPORARY SOLUTION TO HAVE ONLINE GENERATION
            import random
            import string
            rand = "".join(random.choices(string.ascii_lowercase))
            ocp._method.set_name("/codegen/generation_pose_"+rand)
        else:
            ocp.method(rockit.MultipleShooting(N=window_len-1))
            ocp.solver('ipopt', {'expand':True})
            # ocp.solver('ipopt',{"print_time":True,"expand":True},{'tol':1e-4,'print_level':0,'ma57_automatic_scaling':'no','linear_solver':'mumps','max_iter':100})

        
        # Save variables
        self.R_t_x = R_t_x
        self.R_t_y = R_t_y
        self.R_t_z = R_t_z
        self.R_t = R_t
        self.R_r_x = R_r_x
        self.R_r_y = R_r_y
        self.R_r_z = R_r_z
        self.R_r = R_r
        self.p_obj = p_obj
        self.R_obj_x = R_obj_x
        self.R_obj_y = R_obj_y
        self.R_obj_z = R_obj_z
        self.R_obj = R_obj
        self.q = q
        self.nb_joints = nb_joints
        self.U = U
        self.qdot = qdot
        self.U_demo = U_demo
        self.w_invars = w_invars
        self.R_t_start = R_t_start
        self.R_t_end = R_t_end
        self.R_r_start = R_r_start
        self.R_r_end = R_r_end
        self.p_obj_start = p_obj_start
        self.p_obj_end = p_obj_end
        self.R_obj_start = R_obj_start
        self.R_obj_end = R_obj_end
        self.q_lim = q_lim
        self.h = h
        self.window_len = window_len
        self.ocp = ocp
        self.fatrop = fatrop_solver
        
        
    def generate_trajectory(self,U_demo,p_obj_init,R_obj_init,R_t_init,R_r_init,q_init,q_lim,R_t_start,R_r_start,R_t_end,R_r_end,p_obj_start,R_obj_start,p_obj_end,R_obj_end, step_size, U_init = None, w_invars = (10**-3)*np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0]), w_high_start = 1, w_high_end = 0, w_high_invars = (10**-3)*np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0]), w_high_active = 0):
        #%%
        if U_init is None:
            U_init = U_demo

        # Initialize states
        self.ocp.set_initial(self.p_obj, p_obj_init[:self.window_len,:].T)
        self.ocp.set_initial(self.R_obj_x, R_obj_init[:self.window_len,:,0].T) 
        self.ocp.set_initial(self.R_obj_y, R_obj_init[:self.window_len,:,1].T) 
        self.ocp.set_initial(self.R_obj_z, R_obj_init[:self.window_len,:,2].T) 
        self.ocp.set_initial(self.R_t_x, R_t_init[:self.window_len,:,0].T)
        self.ocp.set_initial(self.R_t_y, R_t_init[:self.window_len,:,1].T)
        self.ocp.set_initial(self.R_t_z, R_t_init[:self.window_len,:,2].T)
        self.ocp.set_initial(self.R_r_x, R_r_init[:self.window_len,:,0].T) 
        self.ocp.set_initial(self.R_r_y, R_r_init[:self.window_len,:,1].T) 
        self.ocp.set_initial(self.R_r_z, R_r_init[:self.window_len,:,2].T) 
        self.ocp.set_initial(self.q,q_init.T)
            
        # Initialize controls
        self.ocp.set_initial(self.U,U_init[:-1,:].T)
        self.ocp.set_initial(self.qdot, 0.001*np.ones((self.nb_joints,self.window_len-1)))

        # Set values boundary constraints
        self.ocp.set_value(self.R_t_start,R_t_start)
        self.ocp.set_value(self.R_t_end,R_t_end)
        self.ocp.set_value(self.R_r_start,R_r_start)
        self.ocp.set_value(self.R_r_end,R_r_end)
        self.ocp.set_value(self.p_obj_start,p_obj_start)
        self.ocp.set_value(self.p_obj_end,p_obj_end)
        self.ocp.set_value(self.R_obj_start,R_obj_start)
        self.ocp.set_value(self.R_obj_end,R_obj_end)
        self.ocp.set_value(self.q_lim,q_lim)


        # Set values parameters
        self.ocp.set_value(self.h,step_size)
        self.ocp.set_value(self.U_demo, U_demo.T)   
        weights = np.zeros((len(U_demo),6))
        if w_high_active:
            for i in range(len(U_demo)):
                if i >= w_high_start and i <= w_high_end:
                    weights[i,:] = w_high_invars
                else:
                    weights[i,:] = w_invars
        else:
            for i in range(len(U_demo)):
                weights[i,:] = w_invars
        self.ocp.set_value(self.w_invars, weights.T)  

        # Solve the NLP
        sol = self.ocp.solve()
        if self.fatrop:
            tot_time = self.ocp._method.myOCP.get_stats().time_total
        else:
            tot_time = []
        
        self.sol = sol
                
        # Extract the solved variables
        _,i_r1 = sol.sample(self.U[0],grid='control')
        _,i_r2 = sol.sample(self.U[1],grid='control')
        _,i_r3 = sol.sample(self.U[2],grid='control')
        _,i_t1 = sol.sample(self.U[3],grid='control')
        _,i_t2 = sol.sample(self.U[4],grid='control')
        _,i_t3 = sol.sample(self.U[5],grid='control')
        invariants = np.array((i_r1,i_r2,i_r3,i_t1,i_t2,i_t3)).T
        _,new_trajectory_pos = sol.sample(self.p_obj,grid='control')
        _,new_trajectory_rot = sol.sample(self.R_obj,grid='control')
        _,movingframe_pos = sol.sample(self.R_t,grid='control')
        _,movingframe_rot = sol.sample(self.R_r,grid='control')
        _,joint_val = sol.sample(self.q,grid='control')
        return invariants, new_trajectory_pos, new_trajectory_rot, movingframe_pos, movingframe_rot, tot_time, joint_val


import numpy as np
import casadi as cas
import invariants_py.dynamics_invariants as dynamics
import time
from invariants_py import ocp_helper
import numpy as np

class OCP_calc_pos:

    def __init__(self, window_len = 100, bool_unsigned_invariants = False, rms_error_traj = 10**-2):
       
        opti = cas.Opti() # use OptiStack package from Casadi for easy bookkeeping of variables (no cumbersome indexing)
        
        #%% Decision variables and parameters
        
        # Define system states X (unknown object pose + moving frame pose at every time step)
        p_obj = []
        R_t = []
        X = []
        for k in range(window_len):
            p_obj.append(opti.variable(3,1)) # object position
            R_t.append(opti.variable(3,3)) # Frenet-Serret frame
            X.append(cas.vertcat(cas.vec(R_t[k]), cas.vec(p_obj[k])))

        # Define system controls U (invariants at every time step)
        U = opti.variable(3,window_len-1)

        # Define system parameters P (known values in optimization that need to be set right before solving)
        p_obj_m = [] # measured object positions
        R_t_0 = opti.parameter(3,3) # initial translational Frenet-Serret frame at first sample of window
        for k in range(window_len):
            p_obj_m.append(opti.parameter(3,1)) # object position
        h = opti.parameter(1,1) # step size for integration of dynamic equations
        
        #%% Constraints
        
        # Constrain rotation matrices to be orthogonal (only needed for one timestep, property is propagated by integrator)
        opti.subject_to(ocp_helper.tril_vec(R_t[0].T @ R_t[0] - np.eye(3)) == 0)

        # Dynamics constraints (Multiple shooting)
        integrator = dynamics.define_geom_integrator_tra_FSI_casadi(h)
        for k in range(window_len-1):
            # Integrate current state to obtain next state (next rotation and position)
            Xk_end = integrator(X[k],U[:,k],h)
            
            # Continuity constraint
            opti.subject_to(Xk_end==X[k+1])
            
        # Lower bounds on controls
        if bool_unsigned_invariants:
            opti.subject_to(U[0,:]>=0) # lower bounds on control
            opti.subject_to(U[1,:]>=0) # lower bounds on control

        # Measurement fitting constraint
        trajectory_error = 0
        for k in range(window_len):
            err_pos = p_obj[k] - p_obj_m[k] # position error
            trajectory_error = trajectory_error + cas.dot(err_pos,err_pos)    
        opti.subject_to(trajectory_error/window_len/rms_error_traj**2 < 1)

        # Boundary constraints
        #pti.subject_to(self.p_obj[0] == self.p_obj_m[0]) # Fix first measurement
        #opti.subject_to(self.p_obj[N-1] == self.p_obj_m[N-1]) # Fix last measurement
        #opti.subject_to(U[1,-1] == U[1,-2]) # Last sample has no impact on RMS error

        #%% Objective function

        # Minimize moving frame invariants to deal with singularities and noise
        objective_reg = 0
        for k in range(window_len-1):
            err_abs = U[[1,2],k] # value of moving frame invariants
            objective_reg = objective_reg + cas.dot(err_abs,err_abs) # cost term
        objective = objective_reg/(window_len-1) # normalize with window length

        #%% Solver
        opti.minimize(objective)
        opti.solver('ipopt',{"print_time":True,"expand":True},{'gamma_theta':1e-12,'max_iter':200,'tol':1e-4,'print_level':5,'ma57_automatic_scaling':'no','linear_solver':'mumps'})
        
        # Save variables
        self.R_t = R_t
        self.p_obj = p_obj
        self.U = U
        self.p_obj_m = p_obj_m
        self.R_t_0 = R_t_0
        self.window_len = window_len
        self.opti = opti
        self.first_window = True
        self.h = h
         
    def calculate_invariants_global(self,trajectory_meas,stepsize):
        #%% 

        if trajectory_meas.shape[1] == 3:
            measured_positions = trajectory_meas
        else:
            measured_positions = trajectory_meas[:,:3,3]
        N = self.window_len
        
        # Initialize states
        #TODO  this is not correct yet, ex not perpendicular to ey
        Pdiff = np.diff(measured_positions,axis=0)
        ex = Pdiff / np.linalg.norm(Pdiff,axis=1).reshape(N-1,1)
        ex = np.vstack((ex,[ex[-1,:]]))
        ey = np.tile( np.array((0,0,1)), (N,1) )
        ez = np.array([np.cross(ex[i,:],ey[i,:]) for i in range(N)])
        #Pdiff_cross = np.cross(Pdiff[0:-1],Pdiff[1:])
        #ey = Pdiff_cross / np.linalg.norm(Pdiff_cross,axis=1).reshape(N-2,1)
        
        for k in range(N):
            self.opti.set_initial(self.R_t[k], np.array([ex[k,:], ey[k,:], ez[k,:]]).T ) #construct_init_FS_from_traj(meas_traj.Obj_location)
            self.opti.set_initial(self.p_obj[k], measured_positions[k])
        
        # Initialize controls
        for k in range(N-1):    
            self.opti.set_initial(self.U[:,k], np.array([1,1e-1,1e-12]))
            
        # Set values parameters
        #self.opti.set_value(self.R_t_0, np.eye(3,3))
        for k in range(N):
            self.opti.set_value(self.p_obj_m[k], measured_positions[k])       
        self.opti.set_value(self.h,stepsize)

        # ######################
        # ##  DEBUGGING: check integrator in initial values, time step 0 to 1
        # x0 = cas.vertcat(cas.vec(np.eye(3,3)), cas.vec(measured_positions[0]))
        # u0 = 1e-8*np.ones((3,1))
        # integrator = dynamics.define_geom_integrator_tra_FSI_casadi(self.stepsize)
        # x1 = integrator(x0,u0)
        # print(x1)
        # ######################

        # Solve the NLP
        sol = self.opti.solve_limited()
        self.sol = sol
        
        # Extract the solved variables
        calculate_invariants = sol.value(self.U).T
        calculate_invariants =  np.vstack((calculate_invariants,[calculate_invariants[-1,:]]))
        calculated_trajectory = np.array([sol.value(i) for i in self.p_obj])
        calculated_movingframe = np.array([sol.value(i) for i in self.R_t])
        
        return calculate_invariants, calculated_trajectory, calculated_movingframe

    def calculate_invariants_online(self,trajectory_meas,stepsize,sample_jump):
        #%%
        if self.first_window:
            # Calculate invariants in first window
            invariants, calculated_trajectory, calculated_movingframe = self.calculate_invariants_global(trajectory_meas,stepsize)
            self.first_window = False
            
            
            # Add continuity constraints on first sample
            #self.opti.subject_to( self.R_t[0] == self.R_t_0 )
            #self.opti.subject_to( self.p_obj[0] == self.p_obj_m[0])
            
            return invariants, calculated_trajectory, calculated_movingframe
        else:
            
            if trajectory_meas.shape[1] == 3:
                measured_positions = trajectory_meas
            else:
                measured_positions = trajectory_meas[:,:3,3]
            N = self.window_len
            
            #%% Set values parameters
            #for k in range(1,N):
            #    self.opti.set_value(self.p_obj_m[k], measured_positions[k-1])   
            
            for k in range(0,N):
                    self.opti.set_value(self.p_obj_m[k], measured_positions[k])   
            
            # Set other parameters equal to the measurements in that window
            self.opti.set_value(self.R_t_0, self.sol.value(self.R_t[sample_jump]));
            #self.opti.set_value(self.p_obj_m[0], self.sol.value(self.p_obj[sample_jump]));
            
            self.opti.set_value(self.h,stepsize);
        
            #%% First part of window initialized using results from earlier solution
            # Initialize states
            for k in range(N-sample_jump-1):
                self.opti.set_initial(self.R_t[k], self.sol.value(self.R_t[sample_jump+k]))
                self.opti.set_initial(self.p_obj[k], self.sol.value(self.p_obj[sample_jump+k]));
                
            # Initialize controls
            for k in range(N-sample_jump-1):    
                self.opti.set_initial(self.U[:,k], self.sol.value(self.U[:,sample_jump+k]));
                
            #%% Second part of window initialized uses default initialization
            # Initialize states
            for k in range(N-sample_jump,N):
                self.opti.set_initial(self.R_t[k], self.sol.value(self.R_t[-1]))
                self.opti.set_initial(self.p_obj[k], measured_positions[k-1])
                
            # Initialize controls
            for k in range(N-sample_jump-1,N-1):    
                self.opti.set_initial(self.U[:,k], 1e-3*np.ones((3,1)))

            #print(self.sol.value(self.R_t[-1]))

            #%% Solve the NLP
            sol = self.opti.solve_limited()
            self.sol = sol
            
            # Extract the solved variables
            invariants = sol.value(self.U).T
            invariants =  np.vstack((invariants,[invariants[-1,:]]))
            calculated_trajectory = np.array([sol.value(i) for i in self.p_obj])
            calculated_movingframe = np.array([sol.value(i) for i in self.R_t])
            
            return invariants, calculated_trajectory, calculated_movingframe
        
if __name__ == "__main__":
    import matplotlib.pyplot as plt

    # Example data for measured positions and the stepsize
    measured_positions = np.zeros((100, 3))
    t = np.linspace(0, 1, 100)
    radius = 1
    height = 2
    measured_positions[:, 0] = radius * np.cos(t)
    measured_positions[:, 1] = radius * np.sin(t)
    measured_positions[:, 2] = height * t
    stepsize = t[1]-t[0]

    # Plot the measured positions in a 3D graph
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.plot(measured_positions[:, 0], measured_positions[:, 1], measured_positions[:, 2])
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')
    if plt.get_backend() != 'agg':
        plt.show()
    
    # Test the functionalities of the class
    OCP = OCP_calc_pos(window_len=np.size(measured_positions,0), rms_error_traj=10**-3)

    # Call the calculate_invariants_global function and measure the elapsed time
    start_time = time.time()
    calc_invariants, calc_trajectory, calc_movingframes = OCP.calculate_invariants_global(measured_positions, stepsize)
    elapsed_time = time.time() - start_time

    # # Print the results and elapsed time
    # print("Calculated invariants:")
    print(calc_invariants)
    # print("Calculated Moving Frame:")
    # print(calc_movingframes)
    # print("Calculated Trajectory:")
    # print(calc_trajectory)
    # print("Elapsed Time:", elapsed_time, "seconds")


import numpy as np
import casadi as cas
import rockit
import invariants_py.dynamics_invariants as dynamics
import time
from invariants_py.initialization import generate_initvals_from_bounds
from invariants_py.ocp_helper import tril_vec, tril_vec_no_diag, check_solver

class OCP_gen_pos:

    def __init__(self, boundary_constraints, window_len = 100, bool_unsigned_invariants = False, fatrop_solver = False):

        fatrop_solver = check_solver(fatrop_solver)               
       
        #%% Create decision variables and parameters for the optimization problem
        
        ocp = rockit.Ocp(T=1.0)
        
        # Define system states X (unknown object pose + moving frame pose at every time step) 
        p_obj = ocp.state(3) # object position
        R_t_x = ocp.state(3,1) # translational Frenet-Serret frame
        R_t_y = ocp.state(3,1) # translational Frenet-Serret frame
        R_t_z = ocp.state(3,1) # translational Frenet-Serret frame
        R_t = cas.horzcat(R_t_x,R_t_y,R_t_z)

        # Define system controls (invariants at every time step)
        U = ocp.control(3)

        # Define system parameters P (known values in optimization that need to be set right before solving)
        h = ocp.parameter(1)
        
        # Boundary values
        if "moving-frame" in boundary_constraints and "initial" in boundary_constraints["moving-frame"]:
            R_t_start = ocp.parameter(3,3)
        if "moving-frame" in boundary_constraints and "final" in boundary_constraints["moving-frame"]:
            R_t_end = ocp.parameter(3,3)
        if "position" in boundary_constraints and "initial" in boundary_constraints["position"]:
            p_obj_start = ocp.parameter(3)
        if "position" in boundary_constraints and "final" in boundary_constraints["position"]:
            p_obj_end = ocp.parameter(3)
        
        U_demo = ocp.parameter(3,grid='control+') # model invariants
        w_invars = ocp.parameter(3,grid='control+') # weights for invariants

        #%% Specifying the constraints
        
        # Constrain moving frame to be orthogonal (only needed for one timestep, property is propagated by integrator)
        ocp.subject_to(ocp.at_t0(tril_vec(R_t.T @ R_t - np.eye(3))==0.))
        
        # Boundary constraints
        if "moving-frame" in boundary_constraints and "initial" in boundary_constraints["moving-frame"]:
            ocp.subject_to(ocp.at_t0(tril_vec_no_diag(R_t.T @ R_t_start - np.eye(3))) == 0.)
        if "moving-frame" in boundary_constraints and "final" in boundary_constraints["moving-frame"]:
            ocp.subject_to(ocp.at_tf(tril_vec_no_diag(R_t.T @ R_t_end - np.eye(3)) == 0.))
        if "position" in boundary_constraints and "initial" in boundary_constraints["position"]:    
            ocp.subject_to(ocp.at_t0(p_obj == p_obj_start))
        if "position" in boundary_constraints and "final" in boundary_constraints["position"]:
            ocp.subject_to(ocp.at_tf(p_obj == p_obj_end))



        # Dynamic constraints
        (R_t_plus1, p_obj_plus1) = dynamics.vector_invariants_position(R_t, p_obj, U, h)

        # Integrate current state to obtain next state (next rotation and position)
        ocp.set_next(p_obj,p_obj_plus1)
        ocp.set_next(R_t_x,R_t_plus1[:,0])
        ocp.set_next(R_t_y,R_t_plus1[:,1])
        ocp.set_next(R_t_z,R_t_plus1[:,2])
            
        # Lower bounds on controls
        if bool_unsigned_invariants:
            ocp.subject_to(U[0,:]>=0) # lower bounds on control
            ocp.subject_to(U[1,:]>=0) # lower bounds on control
            
        #%% Specifying the objective

        # Fitting constraint to remain close to invariant model
        objective = ocp.sum(1/window_len*cas.dot(w_invars*(U - U_demo),w_invars*(U - U_demo)),include_last=True)

        #%% Define solver and save variables
        ocp.add_objective(objective)
        if fatrop_solver:
            ocp.method(rockit.external_method('fatrop' , N=window_len-1))

            import random, string
            ocp._method.set_name("/codegen/generation_position" + random.choice(string.ascii_lowercase))
        else:
            ocp.method(rockit.MultipleShooting(N=window_len-1))
            ocp.solver('ipopt', {'expand':True})
            # ocp.solver('ipopt',{"print_time":True,"expand":True},{'tol':1e-4,'print_level':0,'ma57_automatic_scaling':'no','linear_solver':'mumps','max_iter':100})

        # Solve already once with dummy measurements
        ocp.set_initial(R_t_x, np.array([1,0,0]))                 
        ocp.set_initial(R_t_y, np.array([0,1,0]))                
        ocp.set_initial(R_t_z, np.array([0,0,1]))
        ocp.set_initial(U, 0.001+np.zeros((3,window_len)))
        ocp.set_value(h,0.1)
        ocp.set_value(U_demo, 0.001+np.zeros((3,window_len)))
        ocp.set_value(w_invars, 0.001+np.zeros((3,window_len)))
        # Boundary constraints
        if "moving-frame" in boundary_constraints and "initial" in boundary_constraints["moving-frame"]:
            ocp.set_value(R_t_start, np.eye(3))
        if "moving-frame" in boundary_constraints and "final" in boundary_constraints["moving-frame"]:
            ocp.set_value(R_t_end, np.eye(3))
        if "position" in boundary_constraints and "initial" in boundary_constraints["position"]:
            ocp.set_value(p_obj_start, np.array([0,0,0]))
        if "position" in boundary_constraints and "final" in boundary_constraints["position"]:
            ocp.set_value(p_obj_end, np.array([1,0,0]))
        ocp.solve_limited()
        
        # OCP to function

        U_model_sampled = ocp.sample(U_demo, grid='control')[1]
        w_sampled = ocp.sample(w_invars, grid='control')[1]

        bounds = []
        bounds_labels = []
        # Boundary constraints
        if "position" in boundary_constraints and "initial" in boundary_constraints["position"]:
            bounds.append(ocp.value(p_obj_start))
            bounds_labels.append("p_obj_start")
        if "position" in boundary_constraints and "final" in boundary_constraints["position"]:
            bounds.append(ocp.value(p_obj_end))
            bounds_labels.append("p_obj_end")
        if "moving-frame" in boundary_constraints and "initial" in boundary_constraints["moving-frame"]:
            bounds.append(ocp.value(R_t_start))
            bounds_labels.append("R_t_start")
        if "moving-frame" in boundary_constraints and "final" in boundary_constraints["moving-frame"]:
            bounds.append(ocp.value(R_t_end))
            bounds_labels.append("R_t_end")

        h_value = ocp.value(h) # value of stepsize

        solution = [ocp.sample(U, grid='control-')[1],
            ocp.sample(p_obj, grid='control')[1], # sampled object positions
            ocp.sample(R_t_x, grid='control')[1], # sampled FS frame (first axis)
            ocp.sample(R_t_y, grid='control')[1], # sampled FS frame (second axis)
            ocp.sample(R_t_z, grid='control')[1]] # sampled FS frame (third axis)
        
        self.first_window = True
        self.ocp = ocp
        self.ocp_function = self.ocp.to_function('ocp_function',
            [h_value,U_model_sampled,w_sampled,*bounds,*solution], # inputs
            [*solution], # outputs
            ["h_value","invars_model","weights",*bounds_labels,"invars1","p_obj1","R_t_x1","R_t_y1","R_t_z1"], # input labels for debugging
            ["invars2","p_obj2","R_t_x2","R_t_y2","R_t_z2"], # output labels for debugging
        )

        # Save variables
        # Save variables (only needed for old way of trajectory generation)
        self.R_t_x = R_t_x
        self.R_t_y = R_t_y
        self.R_t_z = R_t_z
        self.R_t = R_t
        self.p_obj = p_obj
        self.U = U
        self.U_demo = U_demo
        self.w_invars = w_invars
        if "moving-frame" in boundary_constraints and "initial" in boundary_constraints["moving-frame"]:
            self.R_t_start = R_t_start
        if "moving-frame" in boundary_constraints and "final" in boundary_constraints["moving-frame"]:
            self.R_t_end = R_t_end
        if "position" in boundary_constraints and "initial" in boundary_constraints["position"]:
            self.p_obj_start = p_obj_start
        if "position" in boundary_constraints and "final" in boundary_constraints["position"]:
            self.p_obj_end = p_obj_end
        self.h = h
        self.window_len = window_len
        self.ocp = ocp
        self.sol = None
        self.first_window = True
        self.fatrop = fatrop_solver

        #if fatrop_solver:
            #ocp._method.set_option("print_level",0)
            #ocp._method.set_option("tol",1e-11)

    def generate_trajectory(self, invariant_model, boundary_constraints, step_size, weights_params = {}, initial_values = {}):
        
        N = invariant_model.shape[0]

        # weights_params = {
        # 'w_invars': (10**-3)*np.array([1.0, 1.0, 1.0]),
        # 'w_high_start': 1,
        # 'w_high_end': 0,
        # 'w_high_invars': (10**-3)*np.array([1.0, 1.0, 1.0]),
        # 'w_high_active': 0
        # }
        
        # Get the weights for the invariants or set default values
        w_invars = weights_params.get('w_invars', (10**-3)*np.array([1.0, 1.0, 1.0]))
        w_high_start = weights_params.get('w_high_start', N)
        w_high_end = weights_params.get('w_high_end', N)
        w_high_invars = weights_params.get('w_high_invars', (10**-3)*np.array([1.0, 1.0, 1.0]))
        w_high_active = weights_params.get('w_high_active', 0)

        # Set the weights for the invariants
        w_invars = np.tile(w_invars, (len(invariant_model),1)).T
        if w_high_active:
            w_invars[:, w_high_start:w_high_end+1] = w_high_invars.reshape(-1, 1)

        boundary_values_list = [value for sublist in boundary_constraints.values() for value in sublist.values()]

        if self.first_window and not initial_values:
            self.solution,initvals_dict = generate_initvals_from_bounds(boundary_constraints, np.size(invariant_model,0))
            self.first_window = False
        elif self.first_window:
            self.solution = [initial_values["invariants"][:N-1,:].T, initial_values["trajectory"][:N,:].T, initial_values["moving-frames"][:N,:,0].T, initial_values["moving-frames"][:N,:,1].T, initial_values["moving-frames"][:N,:,2].T]
            self.first_window = False

        #print(self.solution)

        # Call solve function
        #print(boundary_values_list)
        self.solution = self.ocp_function(step_size,invariant_model.T,w_invars,*boundary_values_list,*self.solution)

        # Return the results    
        invars, p_obj_sol, R_t_x_sol, R_t_y_sol, R_t_z_sol,  = self.solution # unpack the results     

        print(p_obj_sol.shape)

        invariants = np.array(invars).T # make a N-1 x 3 array
        invariants = np.vstack((invariants, invariants[-1,:])) # make a N x 3 array by repeating last row
        calculated_trajectory = np.array(p_obj_sol).T # make a N x 3 array
        calculated_movingframe = np.reshape(np.hstack((R_t_x_sol[:], R_t_y_sol[:], R_t_z_sol[:])), (-1,3,3)) # make a N x 3 x 3 array

        # # Solve the NLP
        # sol = self.ocp.solve()
        # if self.fatrop:
        #      tot_time = self.ocp._method.myOCP.get_stats().time_total
        # else:
        #      tot_time = 0 

        print(invariants.shape)
        print(invariant_model.shape)

        return invariants, calculated_trajectory, calculated_movingframe, 0

    def generate_trajectory_OLD(self,invariant_model,initial_values,boundary_constraints, step_size, weights_params):
        
        N = np.size(invariant_model,0)
        
        p_obj_init = initial_values['trajectory']
        R_t_init = initial_values['moving-frames']
        U_init = initial_values['invariants']

        R_t_start = boundary_constraints["moving-frame"]["initial"]
        R_t_end = boundary_constraints["moving-frame"]["final"]
        p_obj_start = boundary_constraints["position"]["initial"]
        p_obj_end = boundary_constraints["position"]["final"]

        # Get the weights for the invariants or set default values
        w_invars = weights_params.get('w_invars', (10**-3)*np.array([1.0, 1.0, 1.0]))
        w_high_start = weights_params.get('w_high_start', 1)
        w_high_end = weights_params.get('w_high_end', 0)
        w_high_invars = weights_params.get('w_high_invars', (10**-3)*np.array([1.0, 1.0, 1.0]))
        w_high_active = weights_params.get('w_high_active', 0)

        # Set the weights for the invariants
        weights = np.tile(w_invars, (len(invariant_model), 1))
        if w_high_active:
            weights[w_high_start:w_high_end+1, :] = w_high_invars
        self.ocp.set_value(self.w_invars, weights.T) 

        # Initialize states
        self.ocp.set_initial(self.p_obj, p_obj_init[:self.window_len,:].T)
        self.ocp.set_initial(self.R_t_x, R_t_init[:self.window_len,:,0].T)
        self.ocp.set_initial(self.R_t_y, R_t_init[:self.window_len,:,1].T)
        self.ocp.set_initial(self.R_t_z, R_t_init[:self.window_len,:,2].T)
            
        # Initialize controls
        self.ocp.set_initial(self.U,U_init[:-1,:].T)

        # Set values boundary constraints
        self.ocp.set_value(self.R_t_start,R_t_start)
        self.ocp.set_value(self.R_t_end,R_t_end)
        self.ocp.set_value(self.p_obj_start,p_obj_start)
        self.ocp.set_value(self.p_obj_end,p_obj_end)

        # Set values parameters
        self.ocp.set_value(self.h,step_size)
        self.ocp.set_value(self.U_demo, invariant_model.T)   

        # Solve the NLP
        sol = self.ocp.solve()
        if self.fatrop:
            tot_time = self.ocp._method.myOCP.get_stats().time_total
        else:
            tot_time = 0 
                
        # Extract the solved variables
        _,i_t1 = sol.sample(self.U[0],grid='control')
        _,i_t2 = sol.sample(self.U[1],grid='control')
        _,i_t3 = sol.sample(self.U[2],grid='control')
        invariants = np.array((i_t1,i_t2,i_t3)).T
        _,calculated_trajectory = sol.sample(self.p_obj,grid='control')
        _,calculated_movingframe = sol.sample(self.R_t,grid='control')
        
        return invariants, calculated_trajectory, calculated_movingframe, tot_time

if __name__ == "__main__":

    # Randomly chosen data
    N = 100
    invariant_model = np.zeros((N,3))

    # Boundary constraints
    boundary_constraints = {
        "position": {
            "initial": np.array([0, 0, 0]),
            "final": np.array([1, 0, 0])
        },
        "moving-frame": {
            "initial": np.eye(3),
            "final": np.eye(3)
        },
    }
    step_size = 0.1

    # Create an instance of OCP_gen_pos
    ocp = OCP_gen_pos(boundary_constraints,fatrop_solver=True, window_len=N)

    # Call the generate_trajectory function
    invariants, calculated_trajectory, calculated_movingframe, solve_time = ocp.generate_trajectory(invariant_model, boundary_constraints, step_size)

    # Print the results
    # print("Invariants:", invariants)
    #print("Calculated Trajectory:", calculated_trajectory)
    # print("Calculated Moving Frame:", calculated_movingframe)

    # Second call to generate_trajectory
    boundary_constraints["position"]["initial"] = np.array([1, 0, 0])
    boundary_constraints["position"]["final"] = np.array([1, 2, 2])
    invariants, calculated_trajectory, calculated_movingframe, solve_time = ocp.generate_trajectory(invariant_model, boundary_constraints, step_size)

    # Print the results
    #print("Invariants:", invariants)
    print("Calculated Trajectory:", calculated_trajectory)
    #print("Calculated Moving Frame:", calculated_movingframe)




"""
Demo showing adaptation of the approach trajectory of a bottle opener towards a bottle
"""

from pathlib import Path
from scipy.spatial.transform import Rotation as R
import invariants_py.tutorial_scripts as tutorial
from invariants_py import data_handler as dh

# Set path to data file
data_location = dh.find_data_path('beer_1.txt')
# Set solver
use_fatrop_solver = True
# Set type of trajectory, options: "pose", "position", "rotation"
traj_type = "pose"
"""
Part 1: calculation invariants
"""
# Calculate invariants
invariants_calculation = tutorial.calculate_invariants(data_location, use_fatrop_solver = use_fatrop_solver, traj_type = traj_type)

"""
Part 2: generation new trajectories
"""
# Define target
target_pos = invariants_calculation.Obj_pos[-1]+[0.1,-0.1,-0.05]
target_rotate = R.from_euler('z', 45, degrees=True) # define rotation transformation for the target
generated_trajectory = tutorial.generate_trajectory(data_location, invariants_calculation, target_pos, target_rotate, use_fatrop_solver, traj_type = traj_type)


"""
Simple test to calculate invariants using rockit specification
"""

import numpy as np
import invariants_py.reparameterization as reparam
import invariants_py.rockit_calculate_vector_invariants_position as OCP
import invariants_py.plotters as plotters
import os
import time
from invariants_py import data_handler as dh

use_fatrop_solver = True  # True = fatrop, False = ipopt

#%% Load data

# load data
data_location = dh.find_data_path("contour_coordinates.out")
position_data = np.loadtxt(data_location, dtype='float')

# reparameterize to arc length
trajectory,time_profile,arclength,nb_samples,stepsize = reparam.reparameterize_positiontrajectory_arclength(position_data)

# plot data
plotters.plot_2D_contour(trajectory)

#%% Calculate invariants first time

# specify optimization problem symbolically
FS_calculation_problem = OCP.OCP_calc_pos(window_len=nb_samples, fatrop_solver = use_fatrop_solver)

for i in range(20):
    start_time = time.time()
    invariants, trajectory_recon, mf = FS_calculation_problem.calculate_invariants(trajectory,stepsize)
    end_time = time.time()
    print('')
    print("solution time [s]: ")
    print(end_time - start_time)

# figures
print(trajectory)
print(trajectory_recon)
#plotters.plot_trajectory_invariants(trajectory,trajectory_recon,arclength,invariants)





"""
Dealing with input/output of data
"""

import invariants_py.data as data_folder
import numpy as np
from scipy.spatial.transform import Rotation
import os
import pandas as pd

def find_data_path(file_name):
    try:
        module_dir = os.path.dirname(data_folder.__file__)
        data_path = os.path.join(module_dir,file_name)
    except FileNotFoundError:
        print(f"File {file_name} not found.")
        return
    return data_path

def save_invariants_to_csv(progress, invariants, file_name):
    """
    Save the progress and invariants into a CSV file.

    Args:
        progress (ndarray): Array of progress values.
        invariants (ndarray): Array of invariant values.
        filename (str): Name of the CSV file to save.

    Returns:
        None
    """
    module_dir = os.path.dirname(data_folder.__file__)
    data_path = os.path.join(module_dir,file_name)
    np.savetxt(data_path, np.hstack((progress.reshape(-1, 1), invariants)), delimiter=",", header="progress,invariant1,invariant2,invariant3", comments="")

def read_pose_trajectory_from_txt(filepath):
    """
    Reads poses from a txt file
    
    The txt file is space-delimited and has 8 elements in each row.
    First element is the timestamp, second-to-fourth elements are the position 
    coordinates, and last four elements are the quaternion coordinates 
    (scalar-last convention).

    Parameters
    ----------
    name : str
        Name of the text file where the data is stored.

    Returns
    -------

    T_all : numpy array [Nx4x4] 
        array with homogeneous transformation matrix [4x4] for each sample
    
    timestamps : numpy array [Nx1]
        array of timestamps 
    """
    
    try: 
        data = np.loadtxt(filepath, dtype='float')
    except IOError:
        print(f"File {filepath} not found.")
        return
    
    N = np.size(data,0)
    
    timestamps = np.zeros(N)
    time_zero = data[0][0]
    T_all = np.zeros((N,4,4))
    T_all[:,3,3] = 1
    
    for i in range(0, N):
        #timestamp
        timestamps[i] = data[i][0]-time_zero
        #position
        T_all[i,0,3] = data[i][1]
        T_all[i,1,3] = data[i][2]
        T_all[i,2,3] = data[i][3]
        #rotation matrix from quaternion
        T_all[i,0:3,0:3] = Rotation.from_quat([data[i][4], data[i][5], data[i][6], data[i][7]]).as_matrix()

    return T_all,timestamps

  

def read_invariants_from_csv(filepath):
    """
    Read a csv file and store the first column in 'progress' and the last three columns in 'invariants'.

    Parameters
    ----------
    filepath : str
        The path to the csv file.

    Returns
    -------
    progress : pandas.Series
        The first column of the csv file.
    invariants : pandas.DataFrame
        The last three columns of the csv file.
    """
    invariants = pd.read_csv(filepath).values

    return invariants

def read_pose_trajectory_from_csv(filepath):
    """
    

    Parameters
    ----------
    filepath : TYPE
        DESCRIPTION.

    Returns
    -------
    None.

    """
    
def read_position_trajectory_from_txt(filepath):
    """
    

    Parameters
    ----------
    filepath : TYPE
        DESCRIPTION.

    Returns
    -------
    None.

    """
    
def read_position_trajectory_from_csv(filepath):
    """
    

    Parameters
    ----------
    filepath : TYPE
        DESCRIPTION.

    Returns
    -------
    None.

    """    







import numpy as np
import math as math

def orthonormalize_rotation( T ):
    """
    Find closest rotation matrix to a given 3x3 matrix.

    Parameters
    ----------
    T : a (4,4) numpy array describing transformation matrix or a (3,3) numpy array describing a general matrix
        
    Returns
    -------
    Orthonormalised input matrix T
    """
    R = T[:3,:3]

    u, s, vt = np.linalg.svd(R)
    Rn = vt.T @ np.diag([1,1,np.linalg.det(vt.T @ u.T)]) @ u.T
    T[:3,:3] = Rn.T

    return T


"""
Note: this module has been deprecated

Test to decouple the trajectory calculation from the ROS communication

@author: Glenn Maes


"""

import numpy as np
import time
from invariants_py import invariant_descriptor_class as invars

class CalculateTrajectory:
    
    def __init__(self, demo_file_location, parameterization, invariants_file_location = None):
        '''Calculate invariants of demonstrated trajectory'''
        
        # Calculate invariants
        descriptor = invars.MotionTrajectory(motionDataFile = demo_file_location, invariantSignatureFile = invariants_file_location, invariantType = parameterization, suppressPlotting=True)
        poses = descriptor.getPosesFromInvariantSignature(descriptor.getInvariantSignature())
        
        self.current_pose_trajectory = poses
        self.current_twist_trajectory = [np.zeros([1,6])]
        self.current_invariants = descriptor.getInvariantsDemo()['U']
        self.shape_descriptor = descriptor
        
        # Initialize progress along trajectory (goes from 0 to 1)
        self.N = len(poses)
        self.localprogress = 0
        self.globalprogress = 0
        self.s_final = 1.0
        self.prog_rate = 0.2
        
    def first_window(self, startpose, endpose):
        '''Generate the first trajectory (+ it initializes structure optimization problem for faster trajectories later)'''
        new_trajectory, new_twists, invariants, solver_time = self.shape_descriptor.generateFirstWindowTrajectory(startPose = startpose, endPose = endpose)
        
        # Save results to Class
        self.current_pose_trajectory = new_trajectory
        self.current_twist_trajectory = new_twists
        self.currentPose = new_trajectory[0]
        self.currentTwist = new_twists[0]
        self.current_invariants = invariants

    def trajectory_generation(self, currentpose, endpose, localprogress):
        self.globalprogress += (1-self.globalprogress)*localprogress # add progress along remainder of path
        startwindow_index = int(self.globalprogress*self.N)
        currentPose_index = int(localprogress*float(len(self.current_pose_trajectory))) # subtract past trajectory

        if (self.globalprogress < self.s_final) and (len(self.current_pose_trajectory) > 15):
            starttime = time.time()
            # Calculate new trajectory
            new_trajectory, new_twists, invariants, solver_time = self.shape_descriptor.generateNextWindowTrajectory(startwindow_index, currentPose_index, startPose = currentpose, startTwist = self.currentTwist, endPose = endpose)
            
            # Store values
            self.current_pose_trajectory = new_trajectory
            self.current_twist_trajectory = new_twists
            self.current_invariants = invariants
            
            # Set-up new values
            #L = invariants[3,0]
            endtime = time.time()
            l = self.prog_rate * (endtime - starttime)
            self.s_final = 1-l


    def loop_trajectory_generation(self, currentpose, endpose, localprogress):
        '''Generate trajectories towards target'''
        counter = 1 # keep track how many trajectories were calculated already
        globalprogress = 0
        
        while (not globalprogress >= self.s_final) and len(self.current_pose_trajectory) > 15 and not counter == 100:
            starttime = time.time()
            
            globalprogress += (1-globalprogress)*localprogress # add progress along remainder of path
            startwindow_index = int(globalprogress*self.N)
            currentPose_index = int(localprogress*float(len(self.current_pose_trajectory))) # subtract past trajectory
            
            if globalprogress <= self.s_final:
                # Calculate new trajectory
                new_trajectory, new_twists, invariants, solver_time = self.shape_descriptor.generateNextWindowTrajectory(startwindow_index, currentPose_index, startPose = currentpose, startTwist = self.currentTwist, endPose = endpose)
                
                # Store values
                self.current_pose_trajectory = new_trajectory
                self.current_twist_trajectory = new_twists
                self.current_invariants = invariants
                
                # Set up values for next loop
                counter += 1
                
                #L = invariants[3,0]
                endtime = time.time()
                l = self.prog_rate * (endtime-starttime)
                self.s_final = 1-l



r"""
The frames module provides relevant operations on SO3: operations on rotation
matrices and rotational velocities.

* **SO3/so3** : rotation matrices in 3-dof Euclidean space (3x3 orthonormal matrix)
  / rotational velocity (3-vector consisting of 3-dof rotational velocity)

*Erwin Aertbelien, 2021*
"""

#from numba import jit
import numpy as np
from numpy.linalg import norm
from numpy import trace
import math
from math import fabs,copysign,sqrt,atan2,asin

def random():
    """
    Returns a random 4x4 homogeneous transformation matrix
    """
    [U,S,V] = np.linalg.svd(np.random.rand(3,3));
    if np.linalg.det(U) < 0:
        U[:,2]=-U[:,2];
    return U 

def random_traj(N=1):
    """
    Returns a random trajectory of N 4x4 homogeneous transformation matrices
    """
    return np.array([random() for i in range(N)])

def crossmat(v):
    r"""
    Returns a 3x3 skew symmetric matrix corresponding the the 3D vector v.
    (in so3).  (also called the "hat"-operator in so3).
    
    """
    return np.array([ [0, -v[2], v[1]], 
                      [v[2], 0, -v[0]],
                      [-v[1],v[0],0] ])
def crossvec(M):
    r"""
    Returns the vector corresponding to the skew symmetric matrix in so3
    (also sometimes called the "vee" operator)
    
    Parameters
    ----------
    M : a 3x3 numpy array
        Representing a skew symmetric matrix

    Returns
    -------
    v : a (3,) numpy array
        vector corresponding to this skew symmetric matrix.

    """
    return np.array([ M[2,1]-M[1,2], M[0,2]-M[2,0], M[1,0] - M[0,1] ])/2.0

def rot(rotvec,angle):
    r"""
    Returns a rotation matrix in SO3 corresponding to rotating
    around the (NORMALIZED) vector "rotvec" with a given angle "angle"

    Parameters
    ----------
    rotvec : a (3,) numpy array
        a NORMALIZED direction vector about which you rotate
        
    angle : scalar
        angle around which to rotate

    Returns
    -------
    R :  a (3,3) numpy array
        Rotation matrix
    
    Warning
    -------

        This routine takes a normalized rotation vector as an input.  For performance reasons
        it does not verify this.
    """
    ct = math.cos(angle)
    st = math.sin(angle)
    vt = 1-ct
    m_vt_0=vt*rotvec[0]
    m_vt_1=vt*rotvec[1]
    m_vt_2=vt*rotvec[2]
    m_st_0=rotvec[0]*st
    m_st_1=rotvec[1]*st
    m_st_2=rotvec[2]*st
    m_vt_0_1=m_vt_0*rotvec[1]
    m_vt_0_2=m_vt_0*rotvec[2]
    m_vt_1_2=m_vt_1*rotvec[2]
    return np.array( [
        [ct+m_vt_0*rotvec[0],  -m_st_2+m_vt_0_1, m_st_1+m_vt_0_2],
        [m_st_2+m_vt_0_1, ct+m_vt_1*rotvec[1], -m_st_0+m_vt_1_2,],
        [-m_st_1+m_vt_0_2, m_st_0+m_vt_1_2,   ct+m_vt_2*rotvec[2]]
    ])


def expm(M):
    r"""
    Matrix exponential in so3 of a 3x3 skew symmetric matrix

    Parameters
    ----------
    M : a (3,3) numpy array
        Skew symmetric matrix corresponding to a displacement rotation vector.

    Returns
    -------
    R : a (3,3) numpy array 
        correspondng the resulting orthonormal orientation matrix

    """
    rotvec = crossvec(M)
    n      = norm(rotvec)
    if n==0:
        return np.eye(3)
    else:
        return rot(rotvec/n,n)

def getRot(R):
    r"""
    gets the rotation axis corresponding to an orthonormal matrix.

    Parameters
    ----------
    R : a (3,3) numpy array
        orthonormal rotation matrix
    Returns
    -------
    Rotation axis vector with an amplitude corresponding to the rotation angle
    """
    axis = crossvec(R)
    sa   = norm(axis)
    ca   = (trace(R)-1)/2.0
    if sa==0:
        return axis*0
    else:
        alpha = math.atan2(sa,ca)/sa
        return axis*alpha
    
def logm(R):
    r"""
    Matrix logarithm of an orthonormal matrix

    Parameters
    ----------
    R : a (3,3) numpy array
        orthonormal rotation matrix
    Returns
    -------
    A skew symmetric matrix corresponding to the displacement rotation
    """
    axis = crossvec(R)
    sa   = norm(axis)
    ca   = (trace(R)-1)/2.0
    if ca<-1:
        ca=-1
    if ca>1:
        ca=1
    if sa<1E-17:
        alpha=1/2.0
    else:
        alpha = math.atan2(sa,ca)/sa/2.0   
    return (R-R.T)*alpha

def rotate_x(alpha):
    """Returns orthonormal rotation matrix corresponding to rotating around X by alpha"""
    return rot([1,0,0],alpha)

def rotate_y(alpha):
    """Returns orthonormal rotation matrix corresponding to rotating around Y by alpha"""
    return rot([0,1,0],alpha)

def rotate_z(alpha):
    """Returns orthonormal rotation matrix corresponding to rotating around Z by alpha"""
    return rot([0,0,1],alpha)

def rotate(axis,alpha):
    """Returns orthonormal rotation matrix corresponding to rotating around axis by alpha"""
    axis = axis/norm(axis)
    return rot(axis,alpha)

def RPY(roll,pitch,yaw):
    """
    gives back a rotation matrix from roll, pitch, yaw angles.
    
    Parameters
    ----------
        roll : float
            roll angle
        pitch : float
            pitch angle
        yaw: float
            yaw angle

    RETURNS
    -------
        R : 3x3 numpy array 
            Rotation matrix obtained by rotating with the specified "roll" around X,
            "pitch" around original Y and "yaw"  around original Z.
            This is exactly the same as rotating around "yaw" around the Z-axis, with "pitch"
            around the NEW Y-axis and "roll" around the NEW X-axis.

    notes
    -----
    from my KDL implementation.

    E. Aertbelien sept 2021
    """
    ca1 = math.cos(yaw)
    sa1 = math.sin(yaw)
    cb1 = math.cos(pitch)
    sb1 = math.sin(pitch)
    cc1 = math.cos(roll)
    sc1 = math.sin(roll)
    return np.array([[ca1*cb1,  ca1*sb1*sc1 - sa1*cc1,  ca1*sb1*cc1 + sa1*sc1],
                    [sa1*cb1,   sa1*sb1*sc1 + ca1*cc1,  sa1*sb1*cc1 - ca1*sc1],
                    [-sb1,      cb1*sc1,                cb1*cc1]]);

def getRPY(R):
    """
    Returns the roll-pitch-yaw representation of the given rotation matrix
       
    Parameters
    ----------
        R : 3x3 numpy array representing a rotation matrix

    Returns
    -------
        roll, pitch, yaw : float,float,float
            angles such that the rotation matrix is obtained by rotating with the specified "roll" around X,
            "pitch" around original Y and "yaw"  around original Z.
            This is exactly the same as rotating around "yaw" around the Z-axis, with "pitch"
            around the NEW Y-axis and "roll" around the NEW X-axis.
    
    Notes
    -----
    from my KDL implementation.

    E. Aertbelien sept 2021
    """
    data=R.flatten()
    epsilon=1E-12;
    pitch = atan2(-data[6], sqrt( data[0]*data[0] +data[3]*data[3] )  )
    if ( fabs(pitch) > (np.pi/2.0-epsilon) ):
        yaw = atan2(	-data[1], data[4])
        roll  = 0.0 ;
    else:
        roll  = atan2(data[7], data[8])
        yaw   = atan2(data[3], data[0])
    return roll,pitch,yaw




"""
A function that handles the calculation of the invariants with different parameterizations
"""

import invariants_py.reparameterization as reparam
import invariants_py.opti_calculate_vector_invariants_position_mf as FS_calculation

def calculate_invariants_translation(trajectory, progress_definition="arclength"):

    # different progress definition: {time, default: arclength, arcangle, screwbased}
    # Reparameterize the trajectory based on arclength
    trajectory_geom, arclength, arclength_n, nb_samples, stepsize = reparam.reparameterize_trajectory_arclength(trajectory)

    # Create an instance of the OCP_calc_pos class
    FS_calculation_problem = FS_calculation.OCP_calc_pos(window_len=nb_samples)

    # Calculate the invariants using the global method
    # TODO make a dictionary of the results from which invariants can be extracted
    result = FS_calculation_problem.calculate_invariants_global(trajectory_geom, stepsize=stepsize)
    invariants = result[0]
    trajectory = result[1]
    return invariants, arclength, trajectory




"""

Note: This class has been deprecated.

Implementation of extended Frenet-Serret invariants:
optimization problems for calculating invariants and generating new trajectories

@author: Zeno Gillis, Victor Van Wymeersch, Maxim Vochten

"""

import sys
sys.path.append('../helper_programs')
from invariants_py import plotters
import time
import csv
import numpy as np
import scipy
from matplotlib import pyplot as plt
from scipy.spatial.transform import Rotation as R
from scipy.linalg import logm
import pickle

#
import math

import casadi as cas
import invariants_py.dynamics_invariants as helper

from collections import OrderedDict

invariant_parameters = dict({
                "inv_type": 'timebased', #"geometric" or "timebased" invariants
                "h":        1.0/60.0,  #sample period
                "N_iter":   200,    #Maximum Number of iterations
                "w_pos":    1,      #Weight on the positions
                "w_rot":    1,      #Weight on the rotations
                "w_deriv":  (10**-3)*np.array([1.0, 1.0, 1.0, 10.0, 1.0, 1.0]), #Weight on the derivative term
                "w_abs":    (10**-10)*np.array([1.0, 1.0, 1.0, 1.0])    #Weight on the absolute term
                })

trajectory_generation_parameters = dict({
        'weight_time'   :   np.array([1e1, 1, 1, 5e1, 1, 1]) ,  #ir1,ir2,ir3,it1,it2,it3
        'weight_geo'    :   np.array([5e0, 1, 1, 10e0, 1, 1])    #Ir1,Ir2,Ir3,It1,It2,It3
        })

moving_window_parameters = dict({
        'window_length'   :   10,   #width of the moving window # 10
        'use_window'    :   True,   #use a moving window along the trajectory, recommended for reducing calculation time of long motions
        'weigths_end_constraints' : np.array([1.0, 1.0]), # position, rotation (test: 30, 1)
        'casadi_opti_stack' : cas.Opti(),
        'first_window_solved' : False,
        'previous_sol' : None,
        'invariant_weights' : (10**-6)*np.array([1e2, 1.0, 1.0, 1.0, 1.0, 1.0]) #(test: (10**-7)*np.array([1e4, 1e3, 1e3, 1e1, 1e2, 1e2])
        })

#list of casadi optistack variables
moving_window_variables = dict({
        'p_obj' : [], 
        'R_obj' : [],
        'R_r' : [],
        'R_t' : [],

        'Theta' : None,
        'L' : None,

        'R_t_startwindow' : None,
        'R_r_startwindow' : None,

        'R_obj_startwindow' : None,
        'p_obj_startwindow' : None,

        'R_obj_end' : None,
        'p_obj_end' : None,

        'R_obj_offset' : None,
        'p_obj_offset' : None,

        'U_demo' : None,
        'U' : None
        })

class MotionTrajectory:

    def __init__(self, motionDataFile = None, invariantSignatureFile = None, invariantType = 'timebased', suppressPlotting = False):
        """
        Shape desciptor from given data file, if none is presented a sine_waveoidal motion profile will be generated
            - the data file 'motionDataFile' can either be a .txt file or a .csv file of timestamps with Quaternion pose data
            - A MotionTrajectory is defined as [ [timeStamp0, Pose0], [timeStamp1, Pose1], ... , [timeStampN, PoseN]]
        """
        if motionDataFile == None:
            print("NO MOTION DATAFILE PRESENTED, loading a default linear motion trajectory")
            self.setMotionTrajectory()
        else:
            self.setMotionTrajectoryFromQuaternionFile(motionDataFile)
#        self.saveMotionTrajectory('./data/saved_motion_trajectory.csv')
        self.setSamplePeriod()

        #%% Save the geometric configuration as well
        geometricPositions, geometricRotations, s, theta, vnorm_n, omeganorm_n = self.calculateGeometricFromTimebased()
        self.setGeometricMotionTrajectory(geometricPositions, geometricRotations, s, theta)

        self.path_variable = s
        self.velocityprofile_trans = vnorm_n
        self.velocityprofile_rot = omeganorm_n

        #%% invariant variables
        invariant_parameters['inv_type'] = invariantType
        if not(motionDataFile == None) and invariantType == 'timebased':
            invariant_parameters['h'] = 1.0/60.0  #self.getSamplePeriod() * 10**-9 #timestaps are in ns here..
        elif invariantType == 'geometric':
            invariant_parameters['h'] = 1.0/len(s) #self.getSamplePeriod()
        self.setInvariantParameters()
            
        if invariantSignatureFile == None:
            invariant_signature, invariants_demo = self.calculateInvariantSignature()
            with open('file1.pkl', 'wb') as file:
                pickle.dump(invariant_signature, file)
            with open('file2.pkl', 'wb') as file:
                pickle.dump(invariants_demo, file)
            #self.saveInvariantSignature('../data/saved_invariant_signature.csv')
        else:
            with open('file1.pkl', 'rb') as file:
                invariant_signature = pickle.load(file)
            with open('file2.pkl', 'rb') as file:
                invariants_demo = pickle.load(file)

        self.setInvariantSignature(invariant_signature)
        self.setInvariantsDemo(invariants_demo)
            
            #self.setInvarientSignatureFileFromFile(invariantSignatureFile)
#
#
#        #%% show the MotionTrajectory
        self.suppressPlotting = suppressPlotting
        if not self.suppressPlotting:
            self.showLoadedMotionTrajectory()
            self.showInvariantSignature('Signature Loaded motion')

        #%% setup MPC window approach
        self.setWindowParameters()

    #%% getters

    def getMotionTrajectory(self):

        if self.__motionTrajectory == None:
            print("NO MOTION TRAJECTORY LOADED OR SPECIFIED YET")

        return self.__motionTrajectory

    def getGeometricMotionTrajectory(self):
        return self.__geometricMotionTrajectory

    def getWindowParameters(self):
        return self.__windowParameters

    def getWindowVariables(self):
        return self.__windowVariables

    def getTimeStamps(self):
        motionTrajectory = self.getMotionTrajectory()
        timeStamps = []
        for entry in motionTrajectory:
            timeStamps.append(entry[0])
        return timeStamps

    def getSamplePeriod(self):
        return self.__samplePeriod

    def getCartesianPoses(self):
        motionTrajectory = self.getMotionTrajectory()
        cartesianPoses = []
        for entry in motionTrajectory:
            cartesianPoses.append(entry[1])
        return cartesianPoses

    def getPositions(self):
        cartesianPoses = self.getCartesianPoses()
        positions = []
        for pose in cartesianPoses:
            positions.append(pose[0:3,3])
        return positions

    def getRotations(self):
        cartesianPoses = self.getCartesianPoses()
        rotations = []
        for pose in cartesianPoses:
            rotations.append(pose[0:3,0:3])
        return rotations

    def getInvariantParameters(self):
        return self.__invariantParameters

    def getInvariantSignature(self):
        return self.__invariantSignature

    def getInvariantsDemo(self):
        return self.__invariantsDemo

    def getPosesFromInvariantSignature(self, invariantSignature):

        poses = []

        positions = invariantSignature['p_obj']
        rotations = invariantSignature['R_obj']

        nb_poses = len(positions)

        for i in range(0,nb_poses):
            temp_matrix = np.eye(4)
            temp_matrix[0:3,3] = positions[i]
            temp_matrix[0:3,0:3] = rotations[i]
            poses.append(temp_matrix)

        return poses

    #%% setters

    def setMotionTrajectory(self, motionTrajectory = None):
        #TODO implement
        print("TODO")
        if motionTrajectory == None:
            self.__motionTrajectory = self.generateLinearMotionTrajectory()
        else:
            self.__motionTrajectory = motionTrajectory

    def setMotionTrajectoryFromQuaternionFile(self, motionDataFile):
        self.__motionTrajectory = self.convertQuaternionDataToMotionTrajectory(motionDataFile)

    def setMotionTrajectoryFromTransformFile(self, motionDataFile):
        self.__motionTrajectory = self.convertTransformDataToMotionTrajectory(motionDataFile)


    def setGeometricMotionTrajectory(self, geometricPositions, geometricRotations, s, theta):

        geometricMotionTrajectory = OrderedDict()
        geometricMotionTrajectory['s'] = s
        geometricMotionTrajectory['theta'] = theta
        geometricMotionTrajectory['positions'] = geometricPositions
        geometricMotionTrajectory['rotations'] = geometricRotations
        self.__geometricMotionTrajectory = geometricMotionTrajectory

    def setWindowParameters(self, windowParameters = None):
        if windowParameters == None:
            self.__windowParameters = moving_window_parameters
        else:
            self.__windowParameters = windowParameters

    def setWindowVariables(self, windowVariables = None):
        if windowVariables == None:
            self.__windowVariables = moving_window_variables
        else:
            self.__windowVariables = windowVariables

    def setSamplePeriod(self, period = None):
        if period == None:
            time_stamps = self.getTimeStamps()
            max_time = max(time_stamps)
            nb_stamps = len(time_stamps)
            sample_period = (max_time-time_stamps[0])/nb_stamps
        else:
            sample_period = period
        self.__samplePeriod = sample_period

    def setInvariantParameters(self, parameterDictionary = None):
        if parameterDictionary == None:
            self.__invariantParameters = invariant_parameters
        else:
            self.__invariantParameters = parameterDictionary

    def setInvariantSignature(self, invariantSignature):
        self.__invariantSignature = invariantSignature

    def setInvariantsDemo(self, invariantsDemo):
        #these are all 6 invariant results budled in one array!!
        self.__invariantsDemo = invariantsDemo


    #%% functions for plotting motion profiles and invariant signatures

    def showLoadedMotionTrajectory(self):
        plotters.plotTrajectory(self.getCartesianPoses(), label="motion from file", title = 'Loaded motion profile', m = '--', mark = True)

    def showInvariantSignature(self, title):
        plotters.plotInvariantSignature(self.getInvariantSignature(), title = title)

    def plotMotionTrajectory(self, poses, figure = None, title = 'Motion Trajectory', label = 'dataline', color = 'b', m = '-', mark = True):
        fig, p_list = plotters.plotTrajectory(poses, figure = figure, label=label, title = title, c = color, m = m, mark = mark)
        return fig, p_list

    def plotInvariantSignature(self, invariantSignature, title = 'Invariant signature of motion trajectory'):
        plotters.plotInvariantSignature(invariantSignature, title = title)


    #%% functions for reading/writing motion profiles and invariant signatures from/to .csv files

    def readCsv(self, filepath):

        with open(filepath, 'rb') as csvfile:
           reader = csv.reader(csvfile)
           rows = list(reader)
        return rows

    def convertRowsToFloat(self, rowsNoLabel):

        float_lst = []

        for i in range(0, len(rowsNoLabel)):
            temp_row = []
            for nb in rowsNoLabel[i]:
                temp_row.append(float(nb))
            float_lst.append(temp_row)
        return float_lst


    def convertQuaternionDataToMotionTrajectory(self, filepath):
        motionTrajectory = [] # motionTrajectory exists of [ [time0, Pose0], [time1, Pose1], ... , [timeN, PoseN]]


        if filepath.endswith('.csv'):
            rows = self.readCsv(filepath)
            float_lst = self.convertRowsToFloat(rows[1:])



            for i in range(0, len(float_lst)):

                row = float_lst[i]
                tempMatrix = np.eye(4)
                tempMatrix[0:3,0:3] = R.from_quat([row[4], row[5], row[6], row[7]]).as_matrix()
                tempMatrix[0,3] = row[1]
                tempMatrix[1,3] = row[2]
                tempMatrix[2,3] = row[3]

                time_stamp = row[0]
                pose = tempMatrix

                motionTrajectory.append([time_stamp, pose])
        elif filepath.endswith('.txt'):
            data = np.loadtxt(filepath, dtype='float')
            for i in range(0, len(data)):
                tempMatrix = np.eye(4)
                tempMatrix[0,3] = data[i][1]
                tempMatrix[1,3] = data[i][2]
                tempMatrix[2,3] = data[i][3]

                tempMatrix[0:3,0:3] = R.from_quat([data[i][4], data[i][5], data[i][6], data[i][7]]).as_matrix()

                time_stamp = data[i][0]
                motionTrajectory.append([time_stamp, tempMatrix])
        return motionTrajectory

    def convertTransformDataToMotionTrajectory(self, filePath):

        rows = self.readCsv(filePath)
        float_lst = self.convertRowsToFloat(rows[1:])

        motionTrajectory = [] # motionTrajectory exists of [ [time0, Pose0], [time1, Pose1], ... , [timeN, PoseN]]

        for i in range(0, len(float_lst), 5):

            time_stamp = float_lst[i]

            line0 = float_lst[i+1][1:]
            line1 = float_lst[i+2][1:]
            line2 = float_lst[i+3][1:]
            line3 = float_lst[i+4][1:]

            tempMatrix = np.eye(4)
            tempMatrix[0] = line0
            tempMatrix[1] = line1
            tempMatrix[2] = line2
            tempMatrix[3] = line3

            motionTrajectory.append([time_stamp, tempMatrix])

        return motionTrajectory

    def setInvariantSignatureFromFileCSV(self, filePath):
        #TODO implemet
        print("TODO")

    def saveMotionTrajectory(self, filePath):
        #write away data for each pose to a file
        motionTrajectory = self.getMotionTrajectory()
        titleRow = ['time', 'pose matrix']
        with open(filePath, 'w') as csvFile:
            writer = csv.writer(csvFile)
            writer.writerow(titleRow)
            counter = 1
            for entry in motionTrajectory:
                rows = [[entry[0]]]
#                rows.append(entry[1])
                rows.append([999, entry[1][0][0], entry[1][0][1], entry[1][0][2], entry[1][0][3]])
                rows.append([999, entry[1][1][0], entry[1][1][1], entry[1][1][2], entry[1][1][3]])
                rows.append([999, entry[1][2][0], entry[1][2][1], entry[1][2][2], entry[1][2][3]])
                rows.append([999, entry[1][3][0], entry[1][3][1], entry[1][3][2], entry[1][3][3]])
                writer.writerows(rows)
                counter += 1

    def saveInvariantSignature(self, filePath):
        """
        First rows will be the parameters,
        then row shows what is in each column
        next rows are samples
        """
        with open(filePath, 'w') as csvFile:
            writer = csv.writer(csvFile)

            parameters = self.getInvariantParameters()
            parameterRows = [["Invariant Parameters"]]
            for k, v in parameters.items():
                parameterRows.append([k, v])
            parameterRows.append(["END"])
            writer.writerows(parameterRows)

            invariantDict = self.getInvariantSignature()
            titleRow = []

            for k,v in list(invariantDict.items()):
                if k.startswith('U'):
                    print(k)
                    titleRow.append(k)

            time_stamps = self.getTimeStamps()
            nb_stamps = len(time_stamps)
            writer.writerow(titleRow)

            dataRows = []
            for i in range(0, nb_stamps-1):
                row = []
                for key in titleRow:
                    row.append(invariantDict[key][i])
                dataRows.append(row)

            writer.writerows(dataRows)


    #%% generate motion profiles
    def generatesine_waveoidalMotion(self):
        #TODO  implement
        print("TODO")

    def generateNextPoseLinear(self, startPose = np.eye(4), endPose = np.eye(4), duration = 5.0, cyclePeriod = 0.05):

        """
        Generate the next pose in a trajectory from startPose to endPose

        @startPose endPose = start cartesian pose, 4x4 pose matrix
        @param endPose = destination cartesian pose, 4x4 pose matrix
        """


        start_pos = startPose[0:3,3]
        end_pos = endPose[0:3,3]

        start_rot = startPose[0:3,0:3]
        end_rot = endPose[0:3,0:3]

        nb_of_waypoints = duration/cyclePeriod

        next_pos = np.zeros([int(nb_of_waypoints)+1, 3])
        next_Rot = np.zeros([int(nb_of_waypoints)+1, 3, 3])
        next_poses = np.zeros([int(nb_of_waypoints)+1, 4,4])
#        interm_vel = np.zeros([int(nb_of_waypoints)+1])

        # calculate intermediate positions
        diff_vec = (end_pos - start_pos)
        increment_vec = diff_vec/nb_of_waypoints

        #calculate intermediate rotations
        dt = 1.0/nb_of_waypoints
        omega = np.zeros((1,3))
        del_R = scipy.linalg.logm(np.matmul(np.transpose(start_rot),end_rot))
        omega[0,0:3] = [-del_R[1,2]/dt, del_R[0,2]/dt, -del_R[0,1]/dt]

        omega_norm = np.sqrt(np.sum(omega**2,1))

        cumm_sum = np.cumsum(omega_norm)*dt
        theta = np.concatenate((np.array([0]),cumm_sum))
        theta_n = np.linspace(0,theta[-1],nb_of_waypoints+1)
        j = 0
        skip_rot = False
        if omega_norm[0] == 0: #if not rotation from start to end is made, skip the rot interpolation part! otherwise dividing by 0
            skip_rot = True
        for i in range(2):
            while (theta_n[i] > theta[j]) :
                j = j + 1

            ##POSITION INTERPOLATION
            interpol_pos = start_pos + i*increment_vec
            next_pos[i] = (interpol_pos)

            if not skip_rot:
                ##ROTATION INTERPOLATION
                theta0 = theta[0]
                theta1 = theta[1]
                R0 = start_rot
                R1 = end_rot

                R_temp = np.matmul(R0, scipy.linalg.expm( ((theta_n[i] - theta0)/(theta1 - theta0))*scipy.linalg.logm(np.matmul(np.transpose(R0),R1)) ))
                next_Rot[i] = R_temp
            else:
                next_Rot[i] = start_rot
            next_poses[i] = (np.vstack((np.hstack((next_Rot[i],np.vstack(next_pos[i]))),np.array([0,0,0,1]))))


        return next_poses[1]

    def generateLinearMotionTrajectory(self, startPose = np.eye(4), endPose = np.eye(4), duration = 5.0, cyclePeriod = 0.05):
        """
        Generate a linear trajectory from startPose to endPose

        @startPose endPose = start cartesian pose, 4x4 pose matrix
        @param endPose = destination cartesian pose, 4x4 pose matrix
        """
        motion_trajectory = []
#        if startPose == None:
#            startPose = np.eye(4)
#        if endPose == None:
#            endPose = startPose.copy()
#            endPose[1,3] += 0.2 #moved 20 in +y-direction

#        cartStart = self.getCartPos()
        start_pos = startPose[0:3,3]
        end_pos = endPose[0:3,3]

        start_rot = startPose[0:3,0:3]
        end_rot = endPose[0:3,0:3]

        nb_of_waypoints = duration/cyclePeriod

        interm_pos = np.zeros([int(nb_of_waypoints)+1, 3])
        interm_Rot = np.zeros([int(nb_of_waypoints)+1, 3, 3])
        interm_poses = np.zeros([int(nb_of_waypoints)+1, 4,4])
#        interm_vel = np.zeros([int(nb_of_waypoints)+1])

        # calculate intermediate positions
        diff_vec = (end_pos - start_pos)
        increment_vec = diff_vec/nb_of_waypoints

        #calculate intermediate rotations
        dt = 1.0/nb_of_waypoints
        omega = np.zeros((1,3))
        del_R = scipy.linalg.logm(np.matmul(np.transpose(start_rot),end_rot))
        omega[0,0:3] = [-del_R[1,2]/dt, del_R[0,2]/dt, -del_R[0,1]/dt]

        omega_norm = np.sqrt(np.sum(omega**2,1))

        cumm_sum = np.cumsum(omega_norm)*dt
        theta = np.concatenate((np.array([0]),cumm_sum))
        theta_n = np.linspace(0,theta[-1],nb_of_waypoints+1)
        j = 0
        skip_rot = False
        if omega_norm[0] == 0: #if not rotation from start to end is made, skip the rot interpolation part! otherwise dividing by 0
            skip_rot = True
        for i in range(int(nb_of_waypoints)+1):
            while (theta_n[i] > theta[j]) :
                j = j + 1

            ##POSITION INTERPOLATION
            interpol_pos = start_pos + i*increment_vec
            interm_pos[i] = (interpol_pos)

            if not skip_rot:
                ##ROTATION INTERPOLATION
                theta0 = theta[0]
                theta1 = theta[1]
                R0 = start_rot
                R1 = end_rot

                R_temp = np.matmul(R0, scipy.linalg.expm( ((theta_n[i] - theta0)/(theta1 - theta0))*scipy.linalg.logm(np.matmul(np.transpose(R0),R1)) ))
                interm_Rot[i] = R_temp
            else:
                interm_Rot[i] = start_rot
            interm_poses[i] = (np.vstack((np.hstack((interm_Rot[i],np.vstack(interm_pos[i]))),np.array([0,0,0,1]))))

            time_stamp = i*cyclePeriod
            motion_trajectory.append([time_stamp, interm_poses[i]])

        return motion_trajectory




    #%% SHAPE DESCRIPTOR CALCULATIONS: invariants, motion generation in new contexts, MPC, ...

    def time2geo(self, posMeas, rotMeas, dt):
        """
        Transform original trajectory P(t) and R(t) to a geometric representation
        in P(s) and R(theta).
        """
        N = len(posMeas)
        omega = np.zeros((N-1,3))
        Pdot = []
        for i in range(N-1):
            del_R = scipy.linalg.logm(np.matmul(np.transpose(rotMeas[i]),rotMeas[i+1]))
            omega[i,:] = [-del_R[1,2]/dt, del_R[0,2]/dt, -del_R[0,1]/dt]
            Pdot.append((posMeas[i+1]-posMeas[i])/dt)
        Pdot= np.array(Pdot)

        omega_norm = np.sqrt(np.sum(omega**2,1))
        vnorm = np.sqrt(np.sum(Pdot**2,1))

        cumm_sum = np.cumsum(omega_norm)*dt
        theta = np.concatenate((np.array([0]),cumm_sum))

        cumm_sum = np.cumsum(vnorm)*dt
        s = np.concatenate((np.array([0]),cumm_sum))

        # Interpolate positions
        p_n = np.linspace(0,s[-1],N)

        x_pos = []
        y_pos = []
        z_pos = []
        p_geo = []
        for pos in posMeas:
            x_pos.append(pos[0])
            y_pos.append(pos[1])
            z_pos.append(pos[2])

        pos_geo_x = np.interp(p_n,s,x_pos)
        pos_geo_y = np.interp(p_n,s,y_pos)
        pos_geo_z = np.interp(p_n,s,z_pos)
          
        p_geo = np.stack((pos_geo_x, pos_geo_y, pos_geo_z), axis=-1)
        skip_rot = False
        # Interpolate rotation matrices
        if omega_norm[0] == 0: #if not rotation from start to end is made, skip the rot interpolation part! otherwise dividing by 0
            skip_rot = True
        theta_n = np.linspace(0,theta[-1],N)
        j = 0
        R_geo = []
        for i in range(len(theta_n)):
            while (theta_n[i] > theta[j+1]) :
                j = j + 1

            if not skip_rot:
                theta0 = theta[j]
                theta1 = theta[j+1]
                R0 = rotMeas[j]
                R1 = rotMeas[j+1]

                R_temp = np.matmul(R0, scipy.linalg.expm( ((theta_n[i] - theta0)/(theta1 - theta0))*scipy.linalg.logm(np.matmul(np.transpose(R0),R1)) ))
                R_geo.append( R_temp)
            else:
                R_geo.append(np.identity(3))
                
        # Dimensionfull velocity profiles, but parameterized in the path variables: v(s) [m/s]
        vnorm_n = np.interp(p_n[0:-1],s[0:-1],vnorm)
        omega_norm_n = np.interp(theta_n[0:-1],theta[0:-1],omega_norm)
                
        return p_geo, R_geo, s, theta, vnorm_n, omega_norm_n

    def calculateGeometricFromTimebased(self):
        # Reparameterise trajectory if requested
        (geometricPositions, geometricRotations, s, theta, vnorm_n, omega_norm_n) = self.time2geo(self.getPositions(), self.getRotations(), self.getSamplePeriod())
        return geometricPositions, geometricRotations, s, theta, vnorm_n, omega_norm_n

    def calculateInvariantSignature(self):

        parameters = self.getInvariantParameters()

        inv_type = parameters["inv_type"]
        h = parameters["h"] #Geometric sample size
        w_pos = parameters["w_pos"]
        w_rot = parameters["w_rot"]
        w_deriv = parameters["w_deriv"]
        w_abs = parameters["w_abs"]

        if parameters['inv_type'] == 'timebased':
            positions = self.getPositions()
            rotations = self.getRotations()
        elif parameters['inv_type'] == 'geometric':
            geoMotionTrajectory = self.getGeometricMotionTrajectory()
            positions = geoMotionTrajectory['positions']
            rotations = geoMotionTrajectory['rotations']
            h = 1.0/len(positions)
            print("USING GEOMETRIC TRAJECTORY TO CALCULATE INVARIANTS")

        N = len(positions)
        ## (Begin) Calculate invariants: ==============================================

        ## Generate optimal eFSI trajectory
        # System states
        R_t  = cas.SX.sym('R_t' ,3,3) # translational Frenet-Serret frame
        R_r = cas.SX.sym('R_r',3,3) # rotational Frenet-Serret frame
        R_obj = cas.SX.sym('R_obj',3,3) # object frame
        p_obj = cas.SX.sym('p_obj',3,1) # object position
        x = cas.vertcat(R_t[:], R_r[:], R_obj[:], p_obj[:])
        #np = length(R_obj(:)) + length(p_obj)

        # System controls (invariants)
        i1 = cas.SX.sym('i1') # object rotation speeds
        i2 = cas.SX.sym('i2') # curvature speed rotational Frenet-Serret
        i3 = cas.SX.sym('i3') # torsion speed rotational Frenet-Serret
        i4 = cas.SX.sym('i4') # object translation speed
        i5 = cas.SX.sym('i5') # curvature speed translational Frenet-Serret
        i6 = cas.SX.sym('i6') # torsion speed translational Frenet-Serret
        u = cas.vertcat(i1, i2, i3, i4, i5, i6)
        nu = np.shape(u)[0] # number of input states

        ## Define a geometric integrator for eFSI, (meaning rigid-body motion is perfectly integrated assuming constant invariants)
        (R_t_plus1, R_r_plus1, R_obj_plus1, p_obj_plus1) = helper.geo_integrator(R_t, R_r, R_obj, p_obj, u, h)
        out_plus1 = cas.vertcat(R_t_plus1[:], R_r_plus1[:], R_obj_plus1[:],  p_obj_plus1)
        integr2 = cas.Function('phi', [x,u] , [out_plus1])


        #===================================================================
        # Build the non-linear optimization problem (NLP) from start to end
        #===================================================================
        opti = cas.Opti()

        # Create variables for multiple shooting method
        p_obj = []
        R_obj = []
        R_r = []
        R_t = []
        X = []

        # System states
        for k in range(N):
            p_obj.append(opti.variable(3,1)) # object position
            R_obj.append(opti.variable(3,3)) # object frame
            R_t.append(opti.variable(3,3)) # translational Frenet-Serret frame
            R_r.append(opti.variable(3,3)) # rotational Frenet-Serret frame
            X.append(cas.vertcat(cas.vec(R_t[k]), cas.vec(R_r[k]), cas.vec(R_obj[k]), cas.vec(p_obj[k])))


        # System controls (invariants)
        U = opti.variable(nu,N-1)
        opti.subject_to(U[0,:]>=0) # lower bounds on control
        opti.subject_to(U[3,:]>=0) # lower bounds on control

        # Using Geometric Invariants?
        if inv_type == "geometric":
            L = opti.variable(1,1) # trajectory total length
            Theta = opti.variable(1,1) # trajectory total angle
            opti.subject_to(L>=0) # lower bounds on L
            opti.subject_to(Theta>=0) # lower bounds Theta
            opti.set_initial(L,1)
            opti.set_initial(Theta,1)
            for k in range(N-1):
                opti.subject_to(U[0,k] == Theta)
                opti.subject_to(U[3,k] == L)

        # Constrain rotation matrices to be orthogonal (only needed for one timestep, property is propagated by integrator)
        opti.subject_to(cas.mtimes(R_t[0].T,R_t[0]) == np.eye(3))
        opti.subject_to(cas.mtimes(R_r[0].T,R_r[0]) == np.eye(3))
        opti.subject_to(cas.mtimes(R_obj[0].T,R_obj[0]) == np.eye(3))

        # Dynamic constraints
        for k in range(N-1):
            # Integrate current state to obtain next state
            #Xk_end = rk4(ode_simp,h,X{k},U(:,k))  #(old integrator)
            Xk_end = integr2(X[k],U[:,k]) #(new integrator)
            # Gap closing constraint
            opti.subject_to(Xk_end==X[k+1])

        # Construct objective
        objective_fit = 0
        for k in range(N):
            err_pos = p_obj[k] - positions[k] # position error
            err_rot = cas.mtimes((rotations[k].T),cas.reshape(R_obj[k],3,3)) - np.eye(3) # rotation error
            objective_fit = objective_fit \
                            + w_pos*cas.dot(err_pos,err_pos) \
                            + w_rot*cas.dot(err_rot,err_rot)

        objective_reg = 0
        for k in range(N-1):
            if k!=0:
                err_deriv = U[:,k] - U[:,k-1] # first-order finite backwards derivative (noise smoothing effect)
            else:
                err_deriv = 0

            err_abs = U[[1,2,4,5],k] # absolute value invariants (force arbitrary invariants to zero)
            ##Check that obj function is correctly typed in !!!
            objective_reg = objective_reg \
                            + cas.dot(w_deriv**(0.5)*err_deriv,w_deriv**(0.5)*err_deriv) \
                            + cas.dot(w_abs**(0.5)*err_abs, w_abs**(0.5)*err_abs)

        objective = objective_fit + objective_reg

        # Initialize states + controls
        for k in range(N):
            opti.set_initial(R_t[k], np.eye(3,3)) #construct_init_FS_from_traj(meas_traj.Obj_location)
            opti.set_initial(R_r[k], np.eye(3,3)) #construct_init_Euler_from_traj(meas_traj.Obj_frames)
            opti.set_initial(p_obj[k], positions[k])
            opti.set_initial(R_obj[k], rotations[k])
            if k!= N-1:
                opti.set_initial(U[:,k], np.ones((6,1)))

        opti.minimize(objective)
        opti.solver('ipopt',{"print_time":True},{'print_level':0,'ma57_automatic_scaling':'no','linear_solver':'mumps'})
        #opti.solver('ipopt',struct(),struct('tol',10e-5))


        ######################
        ##  Debugging stuff

        # Check integrator in initial values, time step 0 to 1
        x0 = cas.vertcat(cas.vec(np.eye(3,3)), cas.vec(np.eye(3,3)), cas.vec(rotations[0]), cas.vec(positions[0]))
        u0 = np.ones((6,1))
        x1 = integr2(x0,u0)
        print(x1)
        ######################

        # Solve the NLP
        sol = opti.solve()
        # (End) Invariant generation =================================================


        # Solution Dictionary
        sol_dict = OrderedDict()
        sol_dict['R_r'] = []
        sol_dict['R_t'] = []
        sol_dict['p_obj'] = []
        sol_dict['R_obj'] = []
        sol_dict['L'] = []
        sol_dict['Theta'] = []

        bundled_invariants = OrderedDict()
        bundled_invariants['U'] = np.zeros((6,N-1))

        for k in range(N): # Extract the generated solutions
            sol_dict['p_obj'].append(sol.value(p_obj[k]))
            sol_dict['R_obj'].append(sol.value(R_obj[k]))
            sol_dict['R_r'].append(sol.value(R_r[k]))
            sol_dict['R_t'].append(sol.value(R_t[k]))
            if k!= N-1:
                bundled_invariants['U'][:,k] = sol.value(U[:,k])

        if inv_type == "geometric":
            sol_dict["L"] = sol.value(L)
            sol_dict["Theta"] = sol.value(Theta)
        else:
            sol_dict["L"].append("not geometric")
            sol_dict["Theta"].append("not geometric")

        bundled_invariants['U'] = np.array(bundled_invariants['U'])
        
        return sol_dict, bundled_invariants

    def generateMotionFromInvariants(self, startPose = np.eye(4), endPose = np.eye(4)):

        #TODO implement GEOMETRIC ONES PROPERLY

        parameters = self.getInvariantParameters()
        inv_type = parameters["inv_type"]
        h = parameters["h"] #Geometric sample size

        if parameters['inv_type'] == 'timebased':
            weights = trajectory_generation_parameters['weight_time']
        elif parameters['inv_type'] == 'geometric':
            weights = trajectory_generation_parameters['weight_geo']

        invariant_signature = self.getInvariantSignature()
        invariants_demo = self.getInvariantsDemo()

        # Initialization

        N = np.shape(invariants_demo['U'])[1]

        if (startPose == np.eye(4)).all():
            p_start = invariant_signature['p_obj'][0]
            R_start = invariant_signature['R_obj'][0]
        else:
            p_start = startPose[0:3,3]
            R_start = startPose[0:3,0:3]

        if (endPose == np.eye(4)).all():
            p_end = invariant_signature['p_obj'][-1]
            R_end = invariant_signature['R_obj'][-1]
        else:
            p_end = endPose[0:3,3]
            R_end = endPose[0:3,0:3]


        ## Generate optimal eFSI trajectory
        # System states
        R_t  = cas.SX.sym('R_t' ,3,3) # translational Frenet-Serret frame
        R_r = cas.SX.sym('R_r',3,3) # rotational Frenet-Serret frame
        R_obj = cas.SX.sym('R_obj',3,3) # object frame
        p_obj = cas.SX.sym('p_obj',3,1) # object position
        x = cas.vertcat(R_t[:], R_r[:], R_obj[:], p_obj[:])


        # System controls (invariants)
        i1 = cas.SX.sym('i1') # object rotation speeds
        i2 = cas.SX.sym('i2') # curvature speed rotational Frenet-Serret
        i3 = cas.SX.sym('i3') # torsion speed rotational Frenet-Serret
        i4 = cas.SX.sym('i4') # object translation speed
        i5 = cas.SX.sym('i5') # curvature speed translational Frenet-Serret
        i6 = cas.SX.sym('i6') # torsion speed translational Frenet-Serret
        u = cas.vertcat(i1, i2, i3, i4, i5, i6)
        nu = np.shape(u)[0] # number of input states


        (R_t_plus1, R_r_plus1, R_obj_plus1, p_obj_plus1) = helper.geo_integrator(R_t, R_r, R_obj, p_obj, u, h)
        out_plus1 = cas.vertcat(R_t_plus1[:], R_r_plus1[:], R_obj_plus1[:],  p_obj_plus1)
        integr2 = cas.Function('phi', [x,u] , [out_plus1])


        # Building the NLP
        opti = cas.Opti()

        # Create variables for multiple shooting method
        p_obj = []
        R_obj = []
        R_r = []
        R_t = []
        X = []

        # System states
        for k in range(N):
            p_obj.append(opti.variable(3,1)) # object position
            R_obj.append(opti.variable(3,3)) # object frame
            R_t.append(opti.variable(3,3)) # translational Frenet-Serret frame
            R_r.append(opti.variable(3,3)) # rotational Frenet-Serret frame
            X.append(cas.vertcat(cas.vec(R_t[k]), cas.vec(R_r[k]), cas.vec(R_obj[k]), cas.vec(p_obj[k])))

        # System controls (invariants)
        U = opti.variable(nu,N-1)

        opti.subject_to(U[0,:]>=0) # lower bounds on control
        opti.subject_to(U[3,:]>=0) # lower bounds on control

        # Using Geometric Invariants?
        if inv_type == "geometric":
            L = opti.variable(1,1) # trajectory total length
            Theta = opti.variable(1,1) # trajectory total angle
            opti.subject_to(L>=0) # lower bounds on L
            opti.subject_to(Theta>=0) # lower bounds Theta
            opti.set_initial(L,1)
            opti.set_initial(Theta,1)
            for k in range(N-1):
                opti.subject_to(U[0,k] == Theta)
                opti.subject_to(U[3,k] == L)

        # Constrain rotation matrices to be orthogonal (only needed for one timestep, property is propagated by integrator)
        opti.subject_to(cas.mtimes(R_t[0].T,R_t[0]) == np.eye(3))
        opti.subject_to(cas.mtimes(R_r[0].T,R_r[0]) == np.eye(3))
        opti.subject_to(cas.mtimes(R_obj[0].T,R_obj[0]) == np.eye(3))

         # Constraints on the start
        opti.subject_to(R_t[0] == invariant_signature['R_t'][0])
        opti.subject_to(R_r[0] == invariant_signature['R_r'][0])
        opti.subject_to(R_obj[0] == R_start)
        opti.subject_to(p_obj[0] == p_start)

        # Constraints on the end
#        opti.subject_to(R_t[-1] == invariant_signature['R_t'][-1])
#        opti.subject_to(R_r[-1] == invariant_signature['R_r'][-1])
        opti.subject_to(R_obj[-1] == R_end)
        opti.subject_to(p_obj[-1] == p_end)


        # Dynamic constraints
        for k in range(N-1):
            # Integrate current state to obtain next state
            #Xk_end = rk4(ode_simp,h,X{k},U(:,k))  #(old integrator)
            Xk_end = integr2(X[k],U[:,k]) #(new integrator)
            # Gap closing constraint
            opti.subject_to(Xk_end==X[k+1])

        # Construct objective
        objective_fit = 0
        for k in range(N-1):
            e = U[:,k] - (invariants_demo['U'][:,k])   # invariants error
            e_weighted = np.sqrt(weights).T * e #mtimes(np.sqrt(weights),e)
            objective_fit = objective_fit + cas.dot(e_weighted,e_weighted) #mtimes(transpose(e_weighted),e_weighted)

        objective = objective_fit

        # Initialize states + controls
        for k in range(N):
            opti.set_initial(R_t[k], invariant_signature['R_t'][k]) #construct_init_FS_from_traj(meas_traj.Obj_location)
            opti.set_initial(R_r[k], invariant_signature['R_r'][k]) #construct_init_Euler_from_traj(meas_traj.Obj_frames)
            opti.set_initial(p_obj[k], invariant_signature['p_obj'][k])
            opti.set_initial(R_obj[k], invariant_signature['R_obj'][k])
            if k!= N-1:
                opti.set_initial(U[:,k], invariants_demo['U'][:,k])

        opti.minimize(objective)
        opti.solver('ipopt')

        sol = opti.solve()


        # Solution Dictionary
        sol_dict = OrderedDict()
        sol_dict['U1'] = []
        sol_dict['U2'] = []
        sol_dict['U3'] = []
        sol_dict['U4'] = []
        sol_dict['U5'] = []
        sol_dict['U6'] = []
        sol_dict['x'] = []
        sol_dict['y'] = []
        sol_dict['z'] = []
        sol_dict['R_r'] = []
        sol_dict['R_t'] = []
        sol_dict['p_obj'] = []
        sol_dict['R_obj'] = []
        sol_dict['L'] = []
        sol_dict['Theta'] = []


        bundled_invariants = OrderedDict()
        bundled_invariants['U'] = np.zeros((6,N-1))

        for k in range(N): # Extract the generated solutions
            sol_dict['p_obj'].append(sol.value(p_obj[k]))
            sol_dict['x'].append(sol_dict['p_obj'][k][0])
            sol_dict['y'].append(sol_dict['p_obj'][k][1])
            sol_dict['z'].append(sol_dict['p_obj'][k][2])
            sol_dict['R_obj'].append(sol.value(R_obj[k]))
            sol_dict['R_r'].append(sol.value(R_r[k]))
            sol_dict['R_t'].append(sol.value(R_t[k]))
            if k!= N-1:
                bundled_invariants['U'][:,k] = sol.value(U[:,k])
                sol_dict['U1'].append(sol.value(U[0,k]))
                sol_dict['U2'].append(sol.value(U[1,k]))
                sol_dict['U3'].append(sol.value(U[2,k]))
                sol_dict['U4'].append(sol.value(U[3,k]))
                sol_dict['U5'].append(sol.value(U[4,k]))
                sol_dict['U6'].append(sol.value(U[5,k]))

        if inv_type == "geometric":
            sol_dict["L"] = sol.value(L)
            sol_dict["Theta"] = sol.value(Theta)
        else:
            sol_dict["L"].append("not geometric")
            sol_dict["Theta"].append("not geometric")

        bundled_invariants['U'] = np.array(bundled_invariants['U'])
        return sol_dict, bundled_invariants


    def generateFirstWindowTrajectory(self, startPose = np.eye(4), startTwist = np.zeros([6,1]), endPose = np.eye(4), simulation = True):
    #def generateFirstWindowTrajectory(self, startPose = np.eye(4), startTwist = np.zeros([6,1]), endPos = np.zeros([3,1]), endRPY = [None, None, None], simulation = True): 
        # RPY test
        #roll_end = cas.SX.sym('roll_end')
        #pitch_end = cas.SX.sym('pitch_end')
        #yaw_end = cas.SX.sym('yaw_end')
        
        ## Weights and paramters
        parameters = self.getInvariantParameters()
        inv_type = parameters["inv_type"]
        h = parameters["h"] #Geometric sample size
        if parameters['inv_type'] == 'timebased':
            weights = trajectory_generation_parameters['weight_time']
        elif parameters['inv_type'] == 'geometric':
            weights = trajectory_generation_parameters['weight_geo']

        invariant_signature = self.getInvariantSignature()
        invariants_demo = self.getInvariantsDemo()

        if (startPose == np.eye(4)).all():
            p_start = invariant_signature['p_obj'][0]
            R_start = invariant_signature['R_obj'][0]
        else:
            print("SETTING STARTPOSE TO GIVEN ONE")
            p_start = startPose[0:3,3]
            R_start = startPose[0:3,0:3]

        if (endPose == np.eye(4)).all():
            p_end = invariant_signature['p_obj'][-1]
            R_end = invariant_signature['R_obj'][-1]
        else:
            print("SETTING ENDPOSE TO GIVEN ONE")
            p_end = endPose[0:3,3]
            R_end = endPose[0:3,0:3]

        #p_end = endPos
            #R_end = [[np.cos(yaw_end)*np.cos(pitch_end), np.cos(yaw_end)*np.sin(pitch_end)*np.sin(roll_end) - np.sin(yaw_end)*np.cos(roll_end), np.cos(yaw_end)*np.sin(pitch_end)*np.cos(roll_end) + np.sin(yaw_end)*np.sin(roll_end)],
            #        [np.sin(yaw_end)*np.cos(pitch_end), np.sin(yaw_end)*np.sin(pitch_end)*np.sin(roll_end) + np.cos(yaw_end)*np.cos(roll_end), np.sin(yaw_end)*np.sin(pitch_end)*np.cos(roll_end) - np.cos(yaw_end)*np.sin(roll_end)],
            #        [-np.sin(pitch_end), np.cos(pitch_end)*np.sin(roll_end), np.cos(pitch_end)*np.cos(roll_end)]]


        ## MPC parameters
        window_parameters = self.getWindowParameters()
        window_length = window_parameters['window_length']
        weights = window_parameters['invariant_weights']
        #weights_end = window_parameters['weigths_end_constraints']

        ## Generate optimal eFSI trajectory
        # System states
        R_t  = cas.SX.sym('R_t',3,3) # translational Frenet-Serret frame
        R_r = cas.SX.sym('R_r',3,3) # rotational Frenet-Serret frame
        R_obj = cas.SX.sym('R_obj',3,3) # object frame
        p_obj = cas.SX.sym('p_obj',3,1) # object position
        x = cas.vertcat(R_t[:], R_r[:], R_obj[:], p_obj[:])
        #np = length(R_obj(:)) + length(p_obj)

        # System controls (invariants)
        i1 = cas.SX.sym('i1') # object rotation speeds
        i2 = cas.SX.sym('i2') # curvature speed rotational Frenet-Serret
        i3 = cas.SX.sym('i3') # torsion speed rotational Frenet-Serret
        i4 = cas.SX.sym('i4') # object translation speed
        i5 = cas.SX.sym('i5') # curvature speed translational Frenet-Serret
        i6 = cas.SX.sym('i6') # torsion speed translational Frenet-Serret
        u = cas.vertcat(i1, i2, i3, i4, i5, i6)
        nu = np.shape(u)[0] # number of input states


        (R_t_plus1, R_r_plus1, R_obj_plus1, p_obj_plus1) = helper.geo_integrator(R_t, R_r, R_obj, p_obj, u, h)
        out_plus1 = cas.vertcat(R_t_plus1[:], R_r_plus1[:], R_obj_plus1[:],  p_obj_plus1)
        integr2 = cas.Function('phi', [x,u] , [out_plus1])


        # Building the NLP
        opti = cas.Opti()

        # Create variables for multiple shooting method
        p_obj = []
        R_obj = []
        R_r = []
        R_t = []
        X = []

        # System states
        for k in range(window_length):
            p_obj.append(opti.variable(3,1)) # object position
            R_obj.append(opti.variable(3,3)) # object frame
            R_t.append(opti.variable(3,3)) # translational Frenet-Serret frame
            R_r.append(opti.variable(3,3)) # rotational Frenet-Serret frame
            X.append(cas.vertcat(cas.vec(R_t[k]), cas.vec(R_r[k]), cas.vec(R_obj[k]), cas.vec(p_obj[k])))

        # System controls (invariants)
        U = opti.variable(nu,window_length-1)
        opti.subject_to(U[0,:]>=0) # lower bounds on control
        opti.subject_to(U[3,:]>=0) # lower bounds on control
        
        # System parameters P (known values, will be set prior to solving)
        R_t_startwindow = opti.parameter(3,3)  # initial FS frame at first sample
        R_r_startwindow = opti.parameter(3,3)  # initial eul frame at first sample
        R_obj_startwindow = opti.parameter(3,3)  # initial obj frame at first sample
        p_obj_startwindow = opti.parameter(3,1)  # initial position at first sample
        
        U_demo = opti.parameter(nu,window_length-1)  # model invariants from demonstrations

        R_obj_end = opti.parameter(3,3)  # target obj frame at final sample
        p_obj_end = opti.parameter(3,1)  # target position at final sample
        R_obj_offset = opti.parameter(3,3)  # offset between end window and target R
        p_obj_offset = opti.parameter(3,1)  # offset between end window and target p

        # TEST
        #rpy_obj_end = opti.parameter(3,1)
        #

        normV = opti.parameter(1,1) 
        v_normalized = opti.parameter(3,1)
        normOmega = opti.parameter(1,1)
        omega_normalized = opti.parameter(3,1)

        # Using Geometric Invariants?
        Theta = None
        L = None
        if inv_type == "geometric":
            L = opti.variable(1,1) # trajectory total length
            Theta = opti.variable(1,1) # trajectory total angle
            opti.subject_to(L>=0) # lower bounds on L
            opti.subject_to(Theta>=0) # lower bounds Theta
            opti.set_initial(L,1)
            opti.set_initial(Theta,1)
            for k in range(window_length-1):
                opti.subject_to(U[0,k] == Theta)
                opti.subject_to(U[3,k] == L)

        # Constrain rotation matrices to be orthogonal (only needed for one timestep, property is propagated by integrator)
        opti.subject_to(cas.mtimes(R_t[0].T,R_t[0]) == np.eye(3))
        opti.subject_to(cas.mtimes(R_r[0].T,R_r[0]) == np.eye(3))
        #opti.subject_to(cas.mtimes(R_obj[0].T,R_obj[0]) == np.eye(3)) # not needed, we constraint R_obj[0] fully

        ## Adding constraint on begin state
        opti.subject_to(R_obj[0] == R_obj_startwindow)
        opti.subject_to(p_obj[0] == p_obj_startwindow)

        opti.subject_to(R_t[0] == R_t_startwindow)
        opti.subject_to(R_r[0] == R_r_startwindow)
        #opti.subject_to(normV*(R_t[0][:,0] - v_normalized) == 0)
        #opti.subject_to(normOmega*(R_r[0][:,0] - omega_normalized) == 0)
        
        
        # Dynamic constraints
        for k in range(window_length-1):
            # Integrate current state to obtain next state
            #Xk_end = rk4(ode_simp,h,X{k},U(:,k))  #(old integrator)
            Xk_end = integr2(X[k],U[:,k]) #(new integrator)
            # Gap closing constraint
            opti.subject_to(Xk_end==X[k+1])

        # Construct objective
        objective = 0
        for k in range(window_length-1):
            e = U[:,k] - U_demo[:,k]  # invariants error
            e_weighted = 1.0*np.sqrt(weights)*e
            objective = objective + cas.mtimes(e_weighted.T,e_weighted)

         # Constraints on the end $$$!Check multiplication of R_obj_reconstruction!$$$
        R_obj_reconstruction = cas.mtimes(cas.mtimes(cas.mtimes(R_r[-1],helper.rodrigues2(Theta*R_obj_offset)),R_r[-1].T),R_obj[-1])  # apply offset to reconstructed trajectory
        p_obj_reconstruction = cas.mtimes(R_t[-1],L*p_obj_offset) + p_obj[-1]  # apply offset to reconstructed trajectory

        R_constraint = cas.mtimes(R_obj_end.T, R_obj_reconstruction)
        
        #TEST GLENN:
        #test = cas.SX.sym('test',3,3)
        rotvec_error = self.getRot(R_constraint) #R_constraint
        
        ## add SOFT end constraints to the objective
        deltaP = p_obj_end - p_obj_reconstruction
        objective = objective + 1.0*(cas.mtimes(deltaP.T,deltaP)) # 30*cas.mtimes...

#        deltaR = (weights_end[3]* R_constraint[0,1] + weights_end[1]*R_constraint[1,2] )
        #TEST GLENN:
        deltaR = (cas.vec(R_constraint)-np.array([[1.0],[0.0],[0.0],[0.0],[1.0],[0.0],[0.0],[0.0],[1.0]]))
        objective = objective + 1.0*cas.mtimes(deltaR.T, deltaR)
        
        #rotvec_error_selection = rotvec_error[0:2]
        #rotvec_error_val = (rotvec_error_selection - np.array([0.0, 0.0]))
        #objective = objective + 2.0*cas.mtimes(rotvec_error_selection.T, rotvec_error_selection)
        #objective = objective + 1.0*(rotvec_error_selection[0] + rotvec_error_selection[1])
        
        #objective = objective + 0.075*Theta**2
        ## objective is done
        opti.minimize(objective)
        opti.solver('ipopt',{"print_time":True},{'print_level':0,'ma57_automatic_scaling':'no','linear_solver':'mumps'})

         # Initialize states + controls
        for k in range(window_length):
            opti.set_initial(R_t[k], invariant_signature['R_t'][k]) #construct_init_FS_from_traj(meas_traj.Obj_location)
            opti.set_initial(R_r[k], invariant_signature['R_r'][k]) #construct_init_Euler_from_traj(meas_traj.Obj_frames)
            opti.set_initial(p_obj[k], invariant_signature['p_obj'][k])
            opti.set_initial(R_obj[k], invariant_signature['R_obj'][k])
            if k!= window_length-1:
                opti.set_initial(U[:,k], invariants_demo['U'][:,k])

        # Set values of parameters that determine the starting constraints
        opti.set_value(R_t_startwindow, invariant_signature['R_t'][0])
        opti.set_value(R_r_startwindow, invariant_signature['R_r'][0])

#        if not simulation:
        opti.set_value(R_obj_startwindow, R_start)
        opti.set_value(p_obj_startwindow, p_start)

        opti.set_value(R_obj_end, R_end)
        
        #TEST
        #for k in range(len(endRPY)):
        #    if endRPY[k] != None:
        #        opti.set_value(rpy_obj_end[k], endRPY[k])
        #R_end = [[np.cos(rpy_obj_end[2])*np.cos(rpy_obj_end[1]), np.cos(rpy_obj_end[2])*np.sin(rpy_obj_end[1])*np.sin(rpy_obj_end[0]) - np.sin(rpy_obj_end[2])*np.cos(rpy_obj_end[0]), np.cos(rpy_obj_end[2])*np.sin(rpy_obj_end[1])*np.cos(rpy_obj_end[0]) + np.sin(rpy_obj_end[2])*np.sin(rpy_obj_end[0])], [np.sin(rpy_obj_end[2])*np.cos(rpy_obj_end[1]), np.sin(rpy_obj_end[2])*np.sin(rpy_obj_end[1])*np.sin(rpy_obj_end[0]) + np.cos(rpy_obj_end[2])*np.cos(rpy_obj_end[0]), np.sin(rpy_obj_end[2])*np.sin(rpy_obj_end[1])*np.cos(rpy_obj_end[0]) - np.cos(rpy_obj_end[2])*np.sin(rpy_obj_end[0])], [-np.sin(rpy_obj_end[1]), np.cos(rpy_obj_end[1])*np.sin(rpy_obj_end[0]), np.cos(rpy_obj_end[1])*np.cos(rpy_obj_end[0])]]

        #opti.subject_to(math.atan2(R_obj_end[2,1], R[2,2]) == rpy_obj_end[0])
        #opti.subject_to(math.atan2(R_obj_end[1,0], R_obj_end[0,0]) == rpy_obj_end[2])
        #opti.subject_to(math.atan2(-R_obj_end[2,0], math.cos(rpy_obj_end[2])*R_obj_end[0,0] + math.sin(rpy_obj_end[2])*R_obj_end[1,0]) == rpy_obj_end[1])
        #opti.subject_to(vec(R_obj_end) == [[np.cos(rpy_obj_end[2])*np.cos(rpy_obj_end[1]), np.cos(rpy_obj_end[2])*np.sin(rpy_obj_end[1])*np.sin(rpy_obj_end[0]) - np.sin(rpy_obj_end[2])*np.cos(rpy_obj_end[0]), np.cos(rpy_obj_end[2])*np.sin(rpy_obj_end[1])*np.cos(rpy_obj_end[0]) + np.sin(rpy_obj_end[2])*np.sin(rpy_obj_end[0])], [np.sin(rpy_obj_end[2])*np.cos(rpy_obj_end[1]), np.sin(rpy_obj_end[2])*np.sin(rpy_obj_end[1])*np.sin(rpy_obj_end[0]) + np.cos(rpy_obj_end[2])*np.cos(rpy_obj_end[0]), np.sin(rpy_obj_end[2])*np.sin(rpy_obj_end[1])*np.cos(rpy_obj_end[0]) - np.cos(rpy_obj_end[2])*np.sin(rpy_obj_end[0])], [-np.sin(rpy_obj_end[1]), np.cos(rpy_obj_end[1])*np.sin(rpy_obj_end[0]), np.cos(rpy_obj_end[1])*np.cos(rpy_obj_end[0])]])
        #opti.subject_to(np.vectorize(R_obj_end) == R_end)
        
        #
        opti.set_value(p_obj_end, p_end)

#        else:
#            #use the ones from the calculation to run the MPC if simulation = TRUE
#            opti.set_value(R_obj_startwindow, invariant_signature['R_obj'][0])
#            opti.set_value(p_obj_startwindow, invariant_signature['p_obj'][0])
#
#            opti.set_value(R_obj_end, invariant_signature['R_obj'][-1])
#            opti.set_value(p_obj_end, invariant_signature['p_obj'][-1])

        (R_offset, p_offset) = helper.offset_integrator(invariants_demo['U'][:, window_length-1:],h)
        opti.set_value(R_obj_offset, logm(R_offset)/invariants_demo['U'][0,0])
        opti.set_value(p_obj_offset, p_offset/invariants_demo['U'][3,0])

        opti.set_value(U_demo[:,:window_length-1], invariants_demo['U'][:, :window_length-1])

        v = startTwist[3:6]
        opti.set_value(normV, cas.norm_2(v))
        if cas.norm_2(v) != 0:
            opti.set_value(v_normalized, v/cas.norm_2(v))
        else:
            opti.set_value(v_normalized, np.array([1,0,0]))
            print('should be first window')
            
        omega = startTwist[0:3]
        opti.set_value(normOmega, cas.norm_2(omega))
        if cas.norm_2(omega) != 0:
            opti.set_value(omega_normalized, omega/cas.norm_2(omega))
        else:
            opti.set_value(omega_normalized, np.array([1,0,0]))
            
            
        sol = opti.solve()
        # TEST
        #print(sol.stats()["t_wall_total"])
        solver_time = sol.stats()["t_wall_total"]
        # retrieve all information, part 1 from window / part 2 by open-loop integration
        
        # Solution Dictionary
        sol_dict = OrderedDict()
        sol_dict['R_r'] = []
        sol_dict['R_t'] = []
        sol_dict['p_obj'] = []
        sol_dict['R_obj'] = []
        sol_dict['T_obj'] = []
        sol_dict['twist_obj'] = []
        sol_dict['L'] = []
        sol_dict['Theta'] = []
        sol_dict['U'] = []
    
        # Part 1: window
        for k in range(window_length): # Extract the generated solutions
            sol_dict['p_obj'].append(sol.value(p_obj[k]))
            sol_dict['R_obj'].append(sol.value(R_obj[k]))
            sol_dict['T_obj'].append(np.r_[np.c_[ sol.value(R_obj[k]),sol.value(p_obj[k]).transpose() ],np.array([[0,0,0,1]])])
            sol_dict['R_r'].append(sol.value(R_r[k]))
            sol_dict['R_t'].append(sol.value(R_t[k]))        
            if k!= window_length-1:
                sol_dict['U'].append(sol.value(U[:,k]))
                omega_obj = sol.value(R_r[k]).dot([sol.value(U[0,k]),0,0]) * self.velocityprofile_rot[k]
                v_obj = sol.value(R_t[k]).dot([sol.value(U[3,k]),0,0]) * self.velocityprofile_trans[k]
                sol_dict['twist_obj'].append(np.append(omega_obj,v_obj))

        if inv_type == "geometric":
            sol_dict["L"] = sol.value(L)
            sol_dict["Theta"] = sol.value(Theta)
        else:
            sol_dict["L"].append("not geometric")
            sol_dict["Theta"].append("not geometric")

        # Part 2: open-loop prediction
        
        # extract end of window
        p_0 = sol.value(p_obj[-1])
        R_t_0 = sol.value(R_t[-1])
        R_r_0 = sol.value(R_r[-1])
        R_obj_0 = sol.value(R_obj[-1])
        
        invariants_demo["U"][3,:] = sol.value(L)
        invariants_demo["U"][0,:] = sol.value(Theta)
        self.setInvariantsDemo(invariants_demo)
        
        invariants_remaining = invariants_demo["U"][:,window_length-1:]
        
        for i in range(1,np.shape(invariants_remaining)[1]+1):
            
            sol_dict['U'].append(invariants_remaining[:,i-1])
            
            omega_0 = np.asarray(R_r_0).dot([invariants_remaining[0,i-1],0,0]) * self.velocityprofile_rot[window_length-1+i-1]
            v_0 = np.asarray(R_t_0).dot([invariants_remaining[3,i-1],0,0]) * self.velocityprofile_trans[window_length-1+i-1]
            sol_dict['twist_obj'].append(np.append(omega_0,v_0))
            
            (R_t_0, R_r_0, R_obj_0, p_0) = helper.geo_integrator_eFSI(R_t_0, R_r_0, R_obj_0, p_0, invariants_remaining[:,i-1], h)
            sol_dict['T_obj'].append(np.r_[np.c_[ R_obj_0 , np.array(p_0) ],np.array([[0,0,0,1]])])

            sol_dict['p_obj'].append(p_0)
            sol_dict['R_obj'].append(R_obj_0)
            sol_dict['R_r'].append(R_r_0)
            sol_dict['R_t'].append(R_t_0)  
                
        ## SAVE the window config so that the next window can be initialised
        window_parameters['first_window_solved'] = True
        window_parameters['casadi_opti_stack'] = opti
        window_parameters['previous_sol'] = sol
        self.setWindowParameters(window_parameters)

        moving_window_variables = dict({
                'p_obj' : p_obj,
                'R_obj' : R_obj,
                'R_r' : R_r,
                'R_t' : R_t,

                'R_t_startwindow' : R_t_startwindow,
                'R_r_startwindow' : R_r_startwindow,

                'R_obj_startwindow' : R_obj_startwindow,
                'p_obj_startwindow' : p_obj_startwindow,

                'R_obj_end' : R_obj_end,
                'p_obj_end' : p_obj_end,

                'R_obj_offset' : R_obj_offset,
                'p_obj_offset' : p_obj_offset,

                'U_demo' : U_demo,

                'Theta' : Theta,
                'L' : L,
                'U' : U,
                
                'v_normalized' : v_normalized,
                'normV' : normV,
                
                'omega_normalized' : omega_normalized,
                'normOmega' : normOmega
                })

        self.setWindowVariables(moving_window_variables)
        
       # windowPoses = self.getPosesFromInvariantSignature(sol_dict)
       # new_trajectory = windowPoses + predicted_poses[1:]
       # window_twist = sol_dict['twist_obj']
        #new_twists = window_twist + predicted_twist
        full_invariants = np.array(sol_dict['U']).transpose()
        
        self.setInvariantSignature(sol_dict)

        return sol_dict['T_obj'], sol_dict['twist_obj'], full_invariants, solver_time

    def generateNextWindowTrajectory(self, n, m, startPose = np.eye(4), startTwist = np.zeros([6,1]), endPose = np.eye(4), simulation = True):

        ## LOAD MPC parameters
        window_parameters = self.getWindowParameters()
        first_window_solved = window_parameters['first_window_solved']
        window_length = window_parameters['window_length']
        #weights_end = window_parameters['weigths_end_constraints']
        opti = window_parameters['casadi_opti_stack']
        previous_sol =  window_parameters['previous_sol']

        if not first_window_solved:
            print("you first have to solve the first window by calling 'generateFirstWindowTrajectory'")
            print("the solutions of that one are necessary to solve the NEXT window")
            return False

        parameters = self.getInvariantParameters()
        inv_type = parameters["inv_type"]
        h = parameters["h"] #Geometric sample size

#        if parameters['inv_type'] == 'timebased':
#            weights = trajectory_generation_parameters['weight_time']
#        elif parameters['inv_type'] == 'geometric':
#            weights = trajectory_generation_parameters['weight_geo']

        invariant_signature = self.getInvariantSignature()
        invariants_demo = self.getInvariantsDemo()

        if (startPose == np.eye(4)).all():
            p_start = invariant_signature['p_obj'][0]
            R_start = invariant_signature['R_obj'][0]
        else:
            p_start = startPose[0:3,3]
            R_start = startPose[0:3,0:3]

        if (endPose == np.eye(4)).all():
            p_end = invariant_signature['p_obj'][-1]
            R_end = invariant_signature['R_obj'][-1]
        else:
            p_end = endPose[0:3,3]
            R_end = endPose[0:3,0:3]

        ## Set the initial values equal to the previous solution
        #opti.set_initial(previous_sol.value_variables())
        lam_g0 = previous_sol.value(opti.lam_g)
        opti.set_initial(opti.lam_g, lam_g0)



        ## LOAD variables!!
        window_variables = self.getWindowVariables()
        p_obj = window_variables['p_obj']
        R_obj = window_variables['R_obj']
        R_r = window_variables['R_r']
        R_t = window_variables['R_t']
        R_t_startwindow = window_variables['R_t_startwindow']
        R_r_startwindow = window_variables['R_r_startwindow']
        R_obj_startwindow = window_variables['R_obj_startwindow']
        p_obj_startwindow = window_variables['p_obj_startwindow']
        R_obj_end = window_variables['R_obj_end']
        p_obj_end = window_variables['p_obj_end']
        R_obj_offset = window_variables['R_obj_offset']
        p_obj_offset = window_variables['p_obj_offset']
        U_demo = window_variables['U_demo']
        normV = window_variables['normV']
        v_normalized = window_variables['v_normalized']
        normOmega = window_variables['normOmega']
        omega_normalized = window_variables['omega_normalized']

        Theta = window_variables['Theta']
        L = window_variables['L']
        U = window_variables['U']


         # Initialize states + controls
        for k in range(window_length):
            opti.set_initial(R_t[k], invariant_signature['R_t'][m+k]) #construct_init_FS_from_traj(meas_traj.Obj_location)
            opti.set_initial(R_r[k], invariant_signature['R_r'][m+k]) #construct_init_Euler_from_traj(meas_traj.Obj_frames)
            opti.set_initial(p_obj[k], invariant_signature['p_obj'][m+k])
            opti.set_initial(R_obj[k], invariant_signature['R_obj'][m+k])
            if k!= window_length-1:
                opti.set_initial(U[:,k], invariants_demo['U'][:,m+k])

        # Set values of parameters that determine the starting constraints
        opti.set_value(R_t_startwindow, invariant_signature['R_t'][m])
        opti.set_value(R_r_startwindow, invariant_signature['R_r'][m])

        p_start = invariant_signature['p_obj'][m]
        R_start = invariant_signature['R_obj'][m]
        p_end = endPose[0:3,3]
        R_end = endPose[0:3,0:3]
            
        ## Set values of parameters that determine the starting constraints
        #opti.set_value(R_t_startwindow, previous_sol.value(R_t[1]));
        #opti.set_value(R_r_startwindow, previous_sol.value(R_r[1]))

        if not simulation:
            opti.set_value(R_obj_startwindow, R_start)
            opti.set_value(p_obj_startwindow, p_start)

            opti.set_value(R_obj_end, R_end)
            opti.set_value(p_obj_end, p_end)

        else:
            #use the ones from the calculation to run the MPC if simulation = TRUE
            opti.set_value(R_obj_startwindow, R_start)
            opti.set_value(p_obj_startwindow, p_start)

            opti.set_value(R_obj_end, R_end)
            opti.set_value(p_obj_end, p_end)

        (R_offset, p_offset) = helper.offset_integrator(invariants_demo['U'][:, n+window_length-1:],h)
        opti.set_value(R_obj_offset, logm(R_offset)/invariants_demo['U'][0,0])
        opti.set_value(p_obj_offset,  p_offset/invariants_demo['U'][3,0])

        opti.set_value(U_demo[:,0:window_length-1], invariants_demo['U'][:, n:n+window_length-1])

        v = startTwist[3:6]
        opti.set_value(normV, cas.norm_2(v))
        if cas.norm_2(v) != 0:
            opti.set_value(v_normalized, v/cas.norm_2(v))
        else:
            opti.set_value(v_normalized, np.array([1,0,0]))
            print('should be first window')
        omega = startTwist[0:3]
        opti.set_value(normOmega, cas.norm_2(omega))
        if cas.norm_2(omega) != 0:
            opti.set_value(omega_normalized, omega/cas.norm_2(omega))
        else:
            opti.set_value(omega_normalized, np.array([1,0,0]))    
        
        sol = opti.solve()
        #sol.stats()
        #print(opti.debug.x_describe(0))
        #print('')
        #print(opti.debug.g_describe(273))
        # TEST
        #print(sol.stats()["t_wall_total"])
        solver_time = sol.stats()["t_wall_total"]

        # retrieve all information, part 1 from window / part 2 by open-loop integration
        
        # Solution Dictionary
        sol_dict = OrderedDict()
        sol_dict['R_r'] = []
        sol_dict['R_t'] = []
        sol_dict['p_obj'] = []
        sol_dict['R_obj'] = []
        sol_dict['T_obj'] = []
        sol_dict['twist_obj'] = []
        sol_dict['L'] = []
        sol_dict['Theta'] = []
        sol_dict['U'] = []

        # Part 1: window
        for k in range(window_length): # Extract the generated solutions
            sol_dict['p_obj'].append(sol.value(p_obj[k]))
            sol_dict['R_obj'].append(sol.value(R_obj[k]))
            sol_dict['T_obj'].append(np.r_[np.c_[ sol.value(R_obj[k]),sol.value(p_obj[k]).transpose() ],np.array([[0,0,0,1]])])
            sol_dict['R_r'].append(sol.value(R_r[k]))
            sol_dict['R_t'].append(sol.value(R_t[k]))
            
            if k!= window_length-1:
                sol_dict['U'].append(sol.value(U[:,k]))
                omega_obj = sol.value(R_r[k]).dot([sol.value(U[0,k]),0,0]) * self.velocityprofile_rot[n+k]
                v_obj = sol.value(R_t[k]).dot([sol.value(U[3,k]),0,0]) * self.velocityprofile_trans[n+k]
                sol_dict['twist_obj'].append(np.append(omega_obj,v_obj))

        if inv_type == "geometric":
            sol_dict["L"] = sol.value(L)
            sol_dict["Theta"] = sol.value(Theta)
        else:
            sol_dict["L"].append("not geometric")
            sol_dict["Theta"].append("not geometric")

        # Part 2: open-loop prediction
        
        # extract end of window
        p_0 = sol.value(p_obj[-1])
        R_t_0 = sol.value(R_t[-1])
        R_r_0 = sol.value(R_r[-1])
        R_obj_0 = sol.value(R_obj[-1])
        
        invariants_demo["U"][3,:] = sol.value(L)
        invariants_demo["U"][0,:] = sol.value(Theta)
        self.setInvariantsDemo(invariants_demo)
        invariants_remaining = invariants_demo["U"][:, n+window_length-1:]
        
        for i in range(1,np.shape(invariants_remaining)[1]+1):
            
            sol_dict['U'].append(invariants_remaining[:,i-1])
            
            omega_0 = np.asarray(R_r_0).dot([invariants_remaining[0,i-1],0,0]) * self.velocityprofile_rot[n+window_length-1+i-1]
            v_0 = np.asarray(R_t_0).dot([invariants_remaining[3,i-1],0,0]) * self.velocityprofile_trans[n+window_length-1+i-1]
            sol_dict['twist_obj'].append(np.append(omega_0,v_0))
            
            (R_t_0, R_r_0, R_obj_0, p_0) = helper.geo_integrator_eFSI(R_t_0, R_r_0, R_obj_0, p_0, invariants_remaining[:,i-1], h)
            sol_dict['T_obj'].append(np.r_[np.c_[ R_obj_0 , np.array(p_0) ],np.array([[0,0,0,1]])])
            sol_dict['p_obj'].append(p_0)
            sol_dict['R_obj'].append(R_obj_0)
            sol_dict['R_r'].append(R_r_0)
            sol_dict['R_t'].append(R_t_0) 


        ## SAVE the window config so that the next window can be initialised
        window_parameters['casadi_opti_stack'] = opti
        window_parameters['previous_sol'] = sol
        self.setWindowParameters(window_parameters)

        moving_window_variables = dict({
                'p_obj' : p_obj,
                'R_obj' : R_obj,
                'R_r' : R_r,
                'R_t' : R_t,

                'R_t_startwindow' : R_t_startwindow,
                'R_r_startwindow' : R_r_startwindow,

                'R_obj_startwindow' : R_obj_startwindow,
                'p_obj_startwindow' : p_obj_startwindow,

                'R_obj_end' : R_obj_end,
                'p_obj_end' : p_obj_end,

                'R_obj_offset' : R_obj_offset,
                'p_obj_offset' : p_obj_offset,

                'U_demo' : U_demo,

                'Theta' : Theta,
                'L' : L,
                'U' : U,
                
                 'normV' : normV,
                 'v_normalized' : v_normalized,
                 
                 'normOmega' : normOmega,
                 'omega_normalized' : omega_normalized

                })

        self.setWindowVariables(moving_window_variables)

        self.setInvariantSignature(sol_dict)
        
        #windowPoses = self.getPosesFromInvariantSignature(sol_dict)
        #new_trajectory = windowPoses + predicted_poses[1:]
        #window_twist = sol_dict['twist_obj']
        #new_twists = window_twist + predicted_twist
        
        #full_invariants = np.append(bundled_invariants['U'].transpose(),invariants.transpose(),axis=0)
        #full_invariants = np.c_[sol_dict['U'],invariants_remaining].transpose()
        full_invariants = np.array(sol_dict['U']).transpose()
        
        return sol_dict['T_obj'], sol_dict['twist_obj'], full_invariants, solver_time
    
    
    def crossvec(self,M):
        return cas.vertcat( (M[2,1]-M[1,2])/2.0, (M[0,2]-M[2,0])/2.0, (M[1,0]-M[0,1])/2.0 )
        #return np.array([ M[2,1]-M[1,2], M[0,2]-M[2,0], M[1,0] - M[0,1] ])/2.0
        #return cas.SX(([ (M[2,1]-M[1,2])/2.0, (M[0,2]-M[2,0])/2.0, (M[1,0]-M[0,1])/2.0 ]))

    def logm_so3(self,R):
        axis = self.crossvec(R)
        sa   = np.linalg.norm(axis)
        ca   = (np.trace(R)-1)/2.0
        if ca<-1:
            ca=-1
        if ca>1:
            ca=1
        if sa<1E-17:
            alpha=1/2.0;
        else:
            alpha = cas.atan2(sa,ca)/sa/2.0   
        return (R-R.T)*alpha
    
    def getRot(self,R):
        axis = self.crossvec(R)
        sa = cas.norm_2(axis)
        ca = (cas.trace(R)-1.0)/2.0
        
        alpha = cas.if_else(sa == 0, 0, cas.atan2(sa,ca)/sa)
        return axis*alpha
        
    
    
    #def getRot(self,R):
        #axis = self.crossvec(R)
        #axis = cas.Function('axis', [R], [ (R[2,1]-R[1,2])/2.0 , (R[0,2]-R[2,0])/2.0, (R[1,0]-R[0,1])/2.0 ])
        #R_test = cas.SX.sym('R_test',3,3)
        #f = cas.Function('f', [R_test], [ (R_test[2,1]-R_test[1,2])/2.0, (R_test[0,2]-R_test[2,0])/2.0, (R_test[1,0]-R_test[0,1])/2.0 ])
        #axis = cas.SX.sym('axis',3)
        #axis = f(R).
        #axis = np.array([ R[2,1]-R[1,2], R[0,2]-R[2,0], R[1,0] - R[0,1] ])/2.0
        #axis = cas.SX([ R[2,1]-R[1,2], R[0,2]-R[2,0], R[1,0] - R[0,1] ]/2.0)
        #axis = cas.SX.sym('axis',3)
        #axis[0] = (R[2,1]-R[1,2])/2.0
        #axis[1] = (R[0,2]-R[2,0])/2.0
        #axis[2] = (R[1,0]-R[0,1])/2.0
        #sa   = np.linalg.norm(axis)
        #ca   = (cas.trace(R)-1)/2.0
        #if sa==0:
        #    return axis*0
        #else:
        #    alpha = math.atan2(sa,ca)/sa
        #alpha = cas.if_else(sa == 0, 0, cas.atan2(sa,ca)/sa)
        #return axis*alpha
        #return axis



if __name__ == "__main(deprecated)__":

    testfile = "/home/roboticskuleuven/catkin_ws/src/invariants_py/data/motion_profiles/single_pose.csv"
    testfile = "/home/roboticskuleuven/catkin_ws/src/invariants_py/data_old/sine_wave.txt"
#    testfile = "/home/roboticskuleuven/catkin_ws/src/invariants_py/data_old/MPC_handover_data.txt"
    testfile = "data/motion_tajectories/recorded_motion.csv"
    descriptor = MotionTrajectory(testfile, invariantType='timebased')
    poses = descriptor.getPosesFromInvariantSignature(descriptor.getInvariantSignature())

    fig, p_lst = descriptor.plotMotionTrajectory(poses, title= "DEMO VS GENERATED trajectory", m = '')



    input("Press enter to generate trajectory")
    new_end_pose = poses[-1].copy()
    new_end_pose[0,3] += 0.01
    new_motion_dict, new_motion_inv = descriptor.generateMotionFromInvariants(endPose=new_end_pose)

    new_poses = descriptor.getPosesFromInvariantSignature(new_motion_dict)

    descriptor.plotMotionTrajectory(new_poses, figure= fig, color='r', label="generated", m='')
#    descriptor.plotInvariantSignature(new_motion_dict, title= "generated trajectory invariant signature")
#
#    equal = True
#    for i in range(0, len(new_motion_dict['p_obj'])):
#        pos1 = new_motion_dict['p_obj'][i]
#        pos2 = descriptor.getInvariantSignature()['p_obj'][i]
#
#        for j in range(0, len(pos1)):
#            if not (round(pos1[j], 3) == round(pos2[j],3)):
#
#                equal = False
#                print pos1[j]
#                print pos2[j]
#                break
#        if equal == False:
#            break
#
#    print 'positions are equal: ' + str(equal)

#
    input("generate first window; Press ENTER")
    window_result = descriptor.generateFirstWindowTrajectory(startPose= poses[0], endPose= new_end_pose)
    windowPoses = descriptor.getPosesFromInvariantSignature(window_result[0])

    first_fig, p_lst = descriptor.plotMotionTrajectory(windowPoses, color='k', title='first window VS demo', label="window", m = '')
    descriptor.plotMotionTrajectory(poses[0:10], figure=first_fig, color='b', label="demo", m = '')
#    descriptor.plotMotionTrajectory(window_result[2], figure= fig, color='g', label="prediction", m ='')
#    descriptor.plotInvariantSignature(window_result[0], title='invariants window trajectory')


#    ## next window
    input("ENTER TO CALCULATE NEXT WINDOW")
    window_result = descriptor.generateNextWindowTrajectory(1)
    windowPoses = descriptor.getPosesFromInvariantSignature(window_result[0])
    descriptor.plotMotionTrajectory(windowPoses, figure= fig, color='m', label="window")

    traveled_poses = []
    for n in range(1, len(descriptor.getPositions())-10):
        window_result = descriptor.generateNextWindowTrajectory(n, startPose=windowPoses[1], endPose= new_end_pose)
        windowPoses = descriptor.getPosesFromInvariantSignature(window_result[0])
        fig, p_lst = descriptor.plotMotionTrajectory(windowPoses, figure= fig, color='m', label="window", mark = False)
        fig, p_lst2 = descriptor.plotMotionTrajectory(window_result[2], figure= fig, color='g', label="prediction", m ='', mark = False)
        traveled_poses.append(windowPoses[0].copy())
#        fig, p_list = descriptor.plotMotionTrajectory(next_windowPoses, figure = fig)
        print(("window nr "+ str(n)))
        plt.pause(0.0001)
        if plt.get_backend() != 'agg':
            plt.show()

        time.sleep(0.02)
        if n < len(descriptor.getPositions())-10:
            plotters.removeMultipleAxis(p_lst)
            plotters.removeMultipleAxis(p_lst2)
#    fig, p_list = descriptor.plotMotionTrajectory(traveled_poses)




import numpy as np
import invariants_py.SO3 as SO3


def interpR(x_n, x_p, R_p):
    """
    Interpolation of a sequence of rotation matrices R_p given at x_p
    to the desired values x_n.
    
    Note: similar syntax as numpy function "interp" for interpolation.

    Parameters
    ----------
    x_n : array [M]
        array of values where rotation matrix must be found
    x_p : array [N]
        array of values where rotation matrix is defined
    R_p : array [Nx3x3]
        sequence of rotation matrices evaluated at x_p

    Returns
    -------
    R_n : array [Mx3x3]
        sequence of rotation matrices evaluated at x_n

        
    TODO check boundaries
        if xq(1) < x(1) || xq(end) > x(end):
            error('Cannot interpolate beyond first or last sample!')
    
    TODO ensure strictly increasing values
        np.all(np.diff(xp) > 0)
    """
    
    j = 0
    R_n = np.zeros((len(x_n),3,3))
    
    for i in range(len(x_n)):
        
        while (x_n[i] > x_p[j+1]):
            j = j+1

        x0 = x_p[j]
        x1 = x_p[j+1]
        R0 = R_p[j]
        R1 = R_p[j+1]

        if x1-x0 > 0:
            R_n[i,:,:] = R0 @ SO3.expm( ((x_n[i]-x0)/(x1-x0)) * SO3.logm(R0.T @ R1) ) 
        else:
            R_n[i,:,:] = R0
            
    return R_n

            
    

def reparameterize_trajectory_arclength(trajectory):
    """
    Reparameterize a given pose trajectory T(t) so that it becomes 
    a function of arc length T(s). The arc length is found from the position
    vector.

    Parameters
    ----------
    trajectory : array [N,4,4]
        Sequence of pose matrices with original parameterization

    Returns
    -------
    trajectory_geom : array [N,4,4]
        Sequence of pose matrices with arc length parameterization
    s : array
        Arc length as a function of original parameterization

    """
    
    N = np.size(trajectory,0)
    
    Rot = trajectory[:,:3,:3]
    P = trajectory[:,:3,3]
    
    # omega = np.zeros((N-1,3))
    # for i in range(N-1):
    #     del_R = SO3.logm( Rot[i].T @ Rot[i+1] ) 
    #     omega[i,:] = SO3.crossvec(del_R)/dt
    # Pdot = np.diff(P)/dt

    # np.linalg.norm(np.diff(P))

    # omega_norm = np.sqrt(np.sum(omega**2,1))
    # vnorm = np.sqrt(np.sum(Pdot**2,1))

    # cumm_sum = np.cumsum(omega_norm)*dt
    # theta = np.concatenate((np.array([0]),cumm_sum))

    # cumm_sum = np.cumsum(vnorm)*dt
    # s = np.concatenate((np.array([0]),cumm_sum))

    Pdiff = np.linalg.norm(np.diff(P,axis=0),axis=1)
    s = np.append(np.zeros(1),np.cumsum(Pdiff))
    s_n = np.linspace(0,s[-1],N)
    
    P_geom = np.array([np.interp(s_n, s, P[:,i]) for i in range(3)]).T
    
    Rot_geom = interpR(s_n, s, Rot)
    
     
    trajectory_geom = trajectory
    trajectory_geom[:,0:3,0:3] = Rot_geom
    trajectory_geom[:,:3,3] = P_geom
        
    # # Dimensionfull velocity profiles, but parameterized in the path variables: v(s) [m/s]
    # vnorm_n = np.interp(p_n[0:-1],s[0:-1],vnorm)
    # omega_norm_n = np.interp(theta_n[0:-1],theta[0:-1],omega_norm)
            
    return trajectory_geom, s, s_n, len(s), 1/len(s) # p_geo, R_geo, s, theta, vnorm_n, omega_norm_n




def reparameterize_positiontrajectory_arclength(trajectory, N=0):
    """
    Reparameterize a given position trajectory p(t) so that it becomes 
    a function of arc length p(s). The arc length is found from the position
    vector.

    Parameters
    ----------
    trajectory : array [N,3]
        Sequence of positions with original parameterization

    Returns
    -------
    trajectory_geom : array [N,3]
        Sequence of positions with arc length parameterization
    s : array
        Arc length as a function of original parameterization

    """
    
    if N==0:
        N = np.size(trajectory,0)
    
    P = trajectory
   
    Pdiff = np.linalg.norm(np.diff(P,axis=0),axis=1)
    s = np.append(np.zeros(1),np.cumsum(Pdiff))
    s_n = np.linspace(0,s[-1],N)
    P_geom = np.array([np.interp(s_n, s, P[:,i]) for i in range(3)]).T
    trajectory_geom = P_geom
       
    return trajectory_geom, s, s_n, len(s_n), 1/len(s)
